--=============================================================================
-- Description: IBIS macro library implemented in VHDL-A(MS)
-- Created: January 2006 by Arpad Muranyi (Intel Corporation)
--
-- Revision history:
--
-- 2006/01/19  - VHDL-A(MS) rev0.0 completed.
-- 2006/02/01  - Added file reader routine to read parameters from a file.
--             - Separated functions into package file.
--             - Changed the type of "i12" and "i21" from real to current to
--               eliminate some warning messages.
-- 2006/02/06  - Fixed a bug in the IBIS buffer models.  "FindCommonLength"
--               function was given the wrong vectors.
--             - Changed the algorithm of "DIV" controlled sources to handle
--               div/0 situation more gently.
--             - Changed the type of "kpd" and "kpu" from real to current to
--               eliminate all warning messages.
--             - Added three more data files to test the IBIS buffer models.
--               These files were extracted with the new Perl script from a
--               real IBIS file, and contain full length IV and Vt tables.
-- 2006/02/27  - Fixed a bug in the "FindCommonLength" and "CommonTime"
--               functions of IBIS buffer models.
--             - Added IV and Vt curve scaling capability to IBIS buffers.
--             - Corrected kC_comp_** initial values and analog equations 
--               for missing C_comp components in Open_*** models.
--             - Removed four unused signals (variables) from IBIS buffers.
-- 2006/06/28  - Fixed "DIV" controlled sources to eliminate div/0 error 
--               due to execution order during first few simulation cycles.
--             - Fixed the input thresholds for buffers and event triggered
--               PWL sources to be more compatible with HSPICE's B-element
--               threshold levels.
--             - Fixed buffers and event controlled PWL sources to initialize
--               properly for all possible combinations of input values. 
--               Also, eliminated possible glitches due to execution order of
--               various simulation cycles.
--             - Added BREAK statements to "DIV" and event controlled PWL
--               sources.
--             - Fixed V-t tables in data files for Open_*** models.
--             - Added a warning to the IBIS_***_DIV elements.
-- 2006/07/05  - Replaced the relational operators in the three IBIS_IO_***
--               models with 'above, and added corresponding break statements.
-- 2006/07/07  - Replaced "<=" with "<" in the event controlled sources when
--               used with the Vth_F or Ith_F parameter to be consistent with
--               the rising edge thresholds and also the buffer models.
-- 2007/03/23  - mikelabonte - Added module documentation. Not tested.
--=============================================================================
--=============================== Module list =================================
--
-- LIST OF MODULES:
--
--   IBIS_R               (p, n)
--   IBIS_VCR             (p, n, ps, ns)
--   IBIS_CCR             (p, n, ps, ns)
--   IBIS_C               (p, n)
--   IBIS_VCC             (p, n, ps, ns)
--   IBIS_CCC             (p, n, ps, ns)
--   IBIS_L               (p, n)
--   IBIS_VCL             (p, n, ps, ns)
--   IBIS_CCL             (p, n, ps, ns)
--   IBIS_K               (p1, n1, p2, n2)
--   IBIS_V               (p, n)
--   IBIS_VCVS            (p, n, ps, ns)
--   IBIS_CCVS            (p, n, ps, ns)
--   IBIS_VCVS_DELAY      (p, n, ps, ns)
--   IBIS_CCVS_DELAY      (p, n, ps, ns)
--   IBIS_VCVS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_ABS        (p, n, ps, ns)
--   IBIS_CCVS_ABS        (p, n, ps, ns)
--   IBIS_VCVS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCVS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCVS_PWL        (p, n, ps, ns)
--   IBIS_CCVS_PWL        (p, n, ps, ns)
--   IBIS_TCVS_PWL        (p, n)
--   IBIS_VECVS_PWL       (p, n, ps, ns)
--   IBIS_CECVS_PWL       (p, n, ps, ns)
--   IBIS_I               (p, n)
--   IBIS_VCCS            (p, n, ps, ns)
--   IBIS_CCCS            (p, n, ps, ns)
--   IBIS_VCCS_DELAY      (p, n, ps, ns)
--   IBIS_CCCS_DELAY      (p, n, ps, ns)
--   IBIS_VCCS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_MIN        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_MAX        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_ABS        (p, n, ps, ns)
--   IBIS_CCCS_ABS        (p, n, ps, ns)
--   IBIS_VCCS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_SUM        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_MULT       (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_CCCS_DIV        (p, n, ps1, ns1, ps2, ns2)
--   IBIS_VCCS_PWL        (p, n, ps, ns)
--   IBIS_CCCS_PWL        (p, n, ps, ns)
--   IBIS_TCCS_PWL        (p, n)
--   IBIS_VECCS_PWL       (p, n, ps, ns)
--   IBIS_CECCS_PWL       (p, n, ps, ns)
--   IBIS_T               (n1, ref1, n2, ref2)
--   IBIS_INPUT           (PC_ref, GC_ref, Pad, Rcv_D)
--   IBIS_OUTPUT          (PU_ref, PD_ref, Pad, In_D, PC_ref, GC_ref)
--   IBIS_IO              (PU_ref, PD_ref, Pad, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--   IBIS_3STATE          (PU_ref, PD_ref, Pad, In_D, En_D, PC_ref, GC_ref)
--   IBIS_OPENSINK        (PU_ref, PD_ref, Pad, In_D, PC_ref, GC_ref)
--   IBIS_IO_OPENSINK     (PU_ref, PD_ref, Pad, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--   IBIS_OPENSOURCE      (PU_ref, PD_ref, Pad, In_D, PC_ref, GC_ref)
--   IBIS_IO_OPENSOURCE   (PU_ref, PD_ref, Pad, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--
--=============================================================================
--
-- MODULE:      IBIS_R (p, n)
--
-- DESCRIPTION:
--              A simple resistor. The resistance value is calculated:
--
--              R = Rval * Scale
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--
-- PARAMETERS:
--
--  Rval        Resistance value
--              DEFAULT Rval = '1.0'
--
--  Scale       Scaling factor for calculating resistance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clk_n clkp IBIS_R Rval=100
--
--  VAMS:       IBIS_R #(.Rval(100)) R1 (clk_n, clk_p);
--
--  VHDL:
--        entity IBIS_R is
--          port (terminal clk_n, clk_p : electrical);
--        end entity IBIS_R;
--        architecture IBIS_SIM of IBIS_R is
--        begin
--          generic map ( Rval => 100 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_R is

    generic (Rval  : real := 1.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_R;
  --===========================================================================
  architecture ideal of IBIS_R is

  quantity Vout across Iout through p to n;

  begin

    Vout == Scale * Rval * Iout;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCR (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled resistor. The resistance value is
--              calculated:
--
--              R = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating resistance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vcrp vcrn IBIS_VCR Scale=50
--
--  VAMS:       IBIS_VCR #(.Scale(50)) VCR1 (clkn, clk_p, vcrp, vcrn);
--
--  VHDL:
--        entity IBIS_VCR is
--          port (terminal clkn, clk_p, vcrp, vcrn : electrical);
--        end entity IBIS_VCR;
--        architecture IBIS_SIM of IBIS_VCR is
--        begin
--          generic map ( Scale => 50 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCR is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCR;
  --===========================================================================
  architecture ideal of IBIS_VCR is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin * Iout;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCR (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled resistor. The resistance value is
--              calculated:
--
--              R = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating resistance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp ccrp ccrn IBIS_CCR Scale=200
--
--  VAMS:       IBIS_CCR #(.Scale(200)) CCR1 (clkn, clk_p, ccrp, ccrn);
--
--  VHDL:
--        entity IBIS_CCR is
--          port (terminal clkn, clk_p, ccrp, ccrn : electrical);
--        end entity IBIS_CCR;
--        architecture IBIS_SIM of IBIS_CCR is
--        begin
--          generic map ( Scale => 200 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCR is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCR;
  --===========================================================================
  architecture ideal of IBIS_CCR is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin * Iout;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_C (p, n)
--
-- DESCRIPTION:
--              A simple capacitor. The capacitance value is calculated:
--
--              C = Cval * Scale
--
-- TERMINALS:
--
--  p           positive capacitor terminal
--  n           negative capacitor terminal
--
-- PARAMETERS:
--
--  Cval        Capacitance value
--              DEFAULT Cval = '1.0'
--
--  V0          Voltage value at time = zero
--              DEFAULT V0 = '0.0'
--
--  Scale       Scaling factor for calculating capacitance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clk_n clkp IBIS_C Cval=4.5 Scale=1e-12
--
--  VAMS:       IBIS_C #(.Cval(4.5),.Scale(1e-12)) C1 (clk_n, clk_p);
--
--  VHDL:
--        entity IBIS_C is
--          port (terminal clk_n, clk_p : electrical);
--        end entity IBIS_C;
--        architecture IBIS_SIM of IBIS_C is
--        begin
--          generic map ( Cval => 4.5, Scale => 1e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_C is

    generic (Cval  : real := 1.0;
             V0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_C;
  --===========================================================================
  architecture ideal of IBIS_C is

  quantity Vout across Iout through p to n;

  begin

    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Scale * Cval * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCC (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled capacitor. The capacitance value is
--              calculated:
--
--              C = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current. This is a
--              charge-conserving model. When the capacitance value
--              changes, a displacement current is produced from p to n to
--              adjust the stored charge so that the total charge level
--              observes the rule:
--
--              Q = C * V(p,n)
--
-- TERMINALS:
--
--  p           positive capacitor terminal
--  n           negative capacitor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  V0          Voltage at time = zero
--              DEFAULT V0 = '0.0'
--
--  Scale       Scaling factor for calculating capacitance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vccp vccn IBIS_VCC Scale=1e-12
--
--  VAMS:       IBIS_VCC #(.Scale(1e-12)) VCC1 (clkn, clk_p, vccp, vccn);
--
--  VHDL:
--        entity IBIS_VCC is
--          port (terminal clkn, clk_p, vccp, vccn : electrical);
--        end entity IBIS_VCC;
--        architecture IBIS_SIM of IBIS_VCC is
--        begin
--          generic map ( Scale => 1e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCC is

    generic (V0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCC;
  --===========================================================================
  architecture ideal of IBIS_VCC is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Vout * Scale * Vin'dot + Scale * Vin * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCC (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled capacitor. The capacitance value is
--              calculated:
--
--              C = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
--              This is a charge-conserving model. When the capacitance
--              value changes, a displacement current is produced from p
--              to n to adjust the stored charge so that the total charge
--              level observes the rule:
--
--              Q = C * V(p,n)
--
-- TERMINALS:
--
--  p           positive capacitor terminal
--  n           negative capacitor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  V0          Capacitor voltage at time zero
--              DEFAULT V0 = '0.0'
--
--  Scale       Scaling factor for calculating capacitance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp cccp cccn IBIS_CCC Scale=5e-12
--
--  VAMS:       IBIS_CCC #(.Scale(5e-12)) CCC1 (clkn, clk_p, cccp, cccn);
--
--  VHDL:
--        entity IBIS_CCC is
--          port (terminal clkn, clk_p, cccp, cccn : electrical);
--        end entity IBIS_CCC;
--        architecture IBIS_SIM of IBIS_CCC is
--        begin
--          generic map ( Scale => 5e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCC is

    generic (V0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCC;
  --===========================================================================
  architecture ideal of IBIS_CCC is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    if domain = quiescent_domain use
      Vout == V0;
    else
      Iout == Vout * Scale * Iin'dot + Scale * Iin * Vout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_L (p, n)
--
-- DESCRIPTION:
--              A simple inductor. The inductance value is calculated:
--
--              L = Lval * Scale
--
-- TERMINALS:
--
--  p           positive inductor terminal
--  n           negative inductor terminal
--
-- PARAMETERS:
--
--  Lval        Inductance value
--              DEFAULT Lval = '1.0'
--
--  I0          Inductance value
--              DEFAULT I0 = '0.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clk_n clkp IBIS_L Lval=0.8 Scale=1e-9
--
--  VAMS:       IBIS_L #(.Lval(0.8),.Scale(1e-9)) L1 (clk_n, clk_p);
--
--  VHDL:
--        entity IBIS_L is
--          port (terminal clk_n, clk_p : electrical);
--        end entity IBIS_L;
--        architecture IBIS_SIM of IBIS_L is
--        begin
--          generic map ( Lval => 0.8, Scale => 1e-9 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_L is

    generic (Lval  : real := 1.0;
             I0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_L;
  --===========================================================================
  architecture ideal of IBIS_L is

  quantity Vout across Iout through p to n;

  begin

    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Scale * Lval * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled inductor. The inductance value is
--              calculated:
--
--              L = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current. This is a
--              flux-conserving model. When the inductance value changes,
--              the voltage-current relationship adjusts so that magnetic
--              flux is conserved in the equation:
--
--              B = L * I(p,n)
--
-- TERMINALS:
--
--  p           positive inductor terminal
--  n           negative inductor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  I0          Inductor current at time zero
--              DEFAULT I0 = '0.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vclp vcln IBIS_VCL Scale=0.5e-9
--
--  VAMS:       IBIS_VCL #(.Scale(0.5e-9)) VCL1 (clkn, clk_p, vclp, vcln);
--
--  VHDL:
--        entity IBIS_VCL is
--          port (terminal clkn, clk_p, vclp, vcln : electrical);
--        end entity IBIS_VCL;
--        architecture IBIS_SIM of IBIS_VCL is
--        begin
--          generic map ( Scale => 0.5e-9 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCL is

    generic (I0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCL;
  --===========================================================================
  architecture ideal of IBIS_VCL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Iout * Scale * Vin'dot + Scale * Vin * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled inductor. The inductance value is
--              calculated:
--
--              L = I(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
--              This is a flux-conserving model. When the inductance value
--              changes, the voltage-current relationship adjusts so that
--              magnetic flux is conserved in the equation:
--
--              B = L * I(p,n)
--
-- TERMINALS:
--
--  p           positive inductor terminal
--  n           negative inductor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  I0          Inductor current at time zero
--              DEFAULT I0 = '0.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp cclp ccln IBIS_CCL Scale=4e-9
--
--  VAMS:       IBIS_CCL #(.Scale(4e-9)) CCL1 (clkn, clk_p, cclp, ccln);
--
--  VHDL:
--        entity IBIS_CCL is
--          port (terminal clkn, clk_p, cclp, ccln : electrical);
--        end entity IBIS_CCL;
--        architecture IBIS_SIM of IBIS_CCL is
--        begin
--          generic map ( Scale => 4e-9 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCL is

    generic (I0    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCL;
  --===========================================================================
  architecture ideal of IBIS_CCL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    if domain = quiescent_domain use
      Iout == I0;
    else
      Vout == Iout * Scale * Iin'dot + Scale * Iin * Iout'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_K (p1, n1, p2, n2)
--
-- DESCRIPTION:
--              A mutually-coupled inductor pair. The inductance values
--              are calculated:
--
--              L_1 = Lval_1 * Scale
--
--              L_2 = Lval_2 * Scale
--
-- TERMINALS:
--
--  p1          positive inductor 1 terminal
--  n1          negative inductor 1 terminal
--  p2          positive inductor 2 terminal
--  n2          negative inductor 2 terminal
--
-- PARAMETERS:
--
--  Lval_1      Inductance value for inductor 1
--              DEFAULT Lval_1 = '1.0'
--
--  Lval_2      Inductance value for inductor 2
--              DEFAULT Lval_2 = '1.0'
--
--  Kval        Mutual coupling value
--              DEFAULT Kval = '0.0'
--
--  I0_1        Inductance value at time zero for inductor 1
--              DEFAULT I0_1 = '0.0'
--
--  I0_2        Inductance value at time zero for inductor 2
--              DEFAULT I0_2 = '0.0'
--
--  Scale       Scaling factor for calculating both inductance values
--              DEFAULT Scale = '1.0'
--
--  M           
--              DEFAULT M = 'Scale * Kval * sqrt(Lval_1 * Lval_2)'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 pad_n pad_p pin_n pin_p IBIS_K \
--              + Lval_1=0.8 Lval_2=0.5 Kval=0.15 Scale=1e-9
--
--  VAMS:       IBIS_K
--              #(.Lval_1(0.8),.Lval_2(0.5),.Kval(0.15),.Scale(1e-9)) \
--              K1 (pad_n, pad_p, pin_n, pin_p);
--
--  VHDL:
--        entity IBIS_K is
--          port (terminal pad_n, pad_p, pin_n, pin_p : electrical);
--        end entity IBIS_K;
--        architecture IBIS_SIM of IBIS_K is
--        begin
--          generic map ( Lval_1 => 0.8, Lval_2 => 0.5,
--          Kval => 0.15, Scale => 1e-9 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_K is

    generic (Lval_1 : real := 1.0;
             Lval_2 : real := 1.0;
             Kval   : real := 0.0;
             I0_1   : real := 0.0;
             I0_2   : real := 0.0;
             Scale  : real := 1.0);

    port (terminal p1, n1, p2, n2 : electrical);

  end entity IBIS_K;
  --===========================================================================
  architecture ideal of IBIS_K is

  quantity Vout1 across Iout1 through p1 to n1;
  quantity Vout2 across Iout2 through p2 to n2;
  constant M : real := Scale * Kval * sqrt(Lval_1 * Lval_2);

  begin

    if domain = quiescent_domain use
      Iout1 == I0_1;
      Iout2 == I0_2;
    else
      Vout1 == Scale * Lval_1 * Iout1'dot + M * Iout2'dot;
      Vout2 == Scale * Lval_2 * Iout2'dot + M * Iout1'dot;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_V (p, n)
--
-- DESCRIPTION:
--              A DC voltage source. Output voltage is calculated:
--
--              V(p,n) = Vdc * Scale
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--
-- PARAMETERS:
--
--  Vdc         DC voltage value
--              DEFAULT Vdc = '1.0'
--
--  Scale       Scaling factor for calculating inductance value
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 vddq vss IBIS_V Vdc=3.3
--
--  VAMS:       IBIS_V #(.Vdc(3.3)) V1 (vddq, vss);
--
--  VHDL:
--        entity IBIS_V is
--          port (terminal vddq, vss : electrical);
--        end entity IBIS_V;
--        architecture IBIS_SIM of IBIS_V is
--        begin
--          generic map ( Vdc => 3.3 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_V is

    generic (Vdc   : real := 1.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_V;
  --===========================================================================
  architecture ideal of IBIS_V is

  quantity Vout across Iout through p to n;

  begin

    Vout == Scale * Vdc;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source. The output voltage
--              value is calculated:
--
--              V(p,n) = V(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 clkn clkp vcrp vcrn IBIS_VCVS Scale=0.5
--
--  VAMS:       IBIS_VCVS #(.Scale(0.5)) VCVS1 (clkn, clk_p, vcrp, vcrn);
--
--  VHDL:
--        entity IBIS_VCVS is
--          port (terminal clkn, clk_p, vcrp, vcrn : electrical);
--        end entity IBIS_VCVS;
--        architecture IBIS_SIM of IBIS_VCVS is
--        begin
--          generic map ( Scale => 0.5 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS;
  --===========================================================================
  architecture ideal of IBIS_VCVS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source. The output voltage
--              value is calculated:
--
--              V(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS Scale=100
--
--  VAMS:       IBIS_CCVS #(.Scale(100)) CCVS1 (buf2in, vss, pad, padi);
--
--  VHDL:
--        entity IBIS_CCVS is
--          port (terminal buf2in, vss, pad, padi : electrical);
--        end entity IBIS_CCVS;
--        architecture IBIS_SIM of IBIS_CCVS is
--        begin
--          generic map ( Scale => 100 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS;
  --===========================================================================
  architecture ideal of IBIS_CCVS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source with delay. All
--              changes at the sense terminals affect the output later in
--              time, using the specified delay. The voltage value is
--              calculated:
--
--              V(p,n) = V(ps,ns) * Scale, delayed by TD
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  TD          Time delay from control voltage change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_VCVS_DELAY TD=200p
--
--  VAMS:       IBIS_VCVS_DELAY #(.TD(200e-12)) \
--              VCVS1 (buf2in, vss, pad, padi);
--
--  VHDL:
--        entity IBIS_VCVS_DELAY is
--          port (terminal buf2in, vss, pad, padi : electrical);
--        end entity IBIS_VCVS_DELAY;
--        architecture IBIS_SIM of IBIS_VCVS_DELAY is
--        begin
--          generic map ( TD => 200e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_VCVS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * Vin'delayed(TD);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source with delay. All
--              changes at the sense terminals affect the output later,
--              using the specified delay in seconds. The voltage value is
--              calculated:
--
--              V(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  TD          Time delay from control current change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS_DELAY Scale=100 TD=200p
--
--  VAMS:       IBIS_CCVS_DELAY #(.Scale(100),.TD(200e-12)) \
--              CCVS1 (buf2in, vss, pad, padi);
--
--  VHDL:
--        entity IBIS_CCVS_DELAY is
--          port (terminal buf2in, vss, pad, padi : electrical);
--        end entity IBIS_CCVS_DELAY;
--        architecture IBIS_SIM of IBIS_CCVS_DELAY is
--        begin
--          generic map ( Scale => 100, TD => 200e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_CCVS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * Iin'delayed(TD);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the lesser
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCVS_MIN
--
--  VAMS:       IBIS_VCVS_MIN #(.Scale(1))
--              MIN1 (bufctl, vss, ctlmain, vss, boost, vss);
--
--  VHDL:
--        entity IBIS_VCVS_MIN is
--          port (terminal bufctl, vss, ctlmain, vss,
--          boost, vss : electrical);
--        end entity IBIS_VCVS_MIN;
--        architecture IBIS_SIM of IBIS_VCVS_MIN is
--        begin
--          generic map ( Scale => 1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCVS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MIN;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * realmin(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the lesser
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain ctlmaini boost boosti IBIS_CCVS_MIN
--
--  VAMS:       IBIS_CCVS_MIN #(.Scale(1)) MIN1
--              (bufctl, vss, ctlmain, vss, boost, vss);
--
--  VHDL:
--        entity IBIS_CCVS_MIN is
--          port (terminal bufctl, vss, ctlmain, vss,
--          boost, vss : electrical);
--        end entity IBIS_CCVS_MIN;
--        architecture IBIS_SIM of IBIS_CCVS_MIN is
--        begin
--          generic map ( Scale => 1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCVS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MIN;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * realmin(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the greater
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCVS_MAX
--
--  VAMS:       IBIS_VCVS_MAX #(.Scale(1)) MAX1
--              (bufctl, vss, ctlmain, vss, boost,vss);
--
--  VHDL:
--        entity IBIS_VCVS_MAX is
--          port (terminal bufctl, vss, ctlmain, vss,
--          boost, vss : electrical);
--        end entity IBIS_VCVS_MAX;
--        architecture IBIS_SIM of IBIS_VCVS_MAX is
--        begin
--          generic map ( Scale => 1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCVS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MAX;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * realmax(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the greater
--              of two input currents. The output voltage value is
--              calculated:
--
--              V(p,n) = max( I(ps1,ns1) I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCVS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MAX;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * realmax(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the
--              absolute value of the control voltage. The output voltage
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( V(ps,ns) ) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_ABS;
  --===========================================================================
  architecture ideal of IBIS_VCVS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Vout == Scale * abs(Vin);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the
--              absolute value of the control current. The output voltage
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( I(ps,ns) ) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_ABS;
  --===========================================================================
  architecture ideal of IBIS_CCVS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Vout == Scale * abs(Iin);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the sum of
--              two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) + V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_SUM;
  --===========================================================================
  architecture ideal of IBIS_VCVS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * (Vin1 + Vin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the sum of
--              two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) + I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_SUM;
  --===========================================================================
  architecture ideal of IBIS_CCVS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * (Iin1 + Iin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the product
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) * V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_MULT;
  --===========================================================================
  architecture ideal of IBIS_VCVS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Vout == Scale * Vin1 * Vin2;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the product
--              of two control input currents. The output voltage value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) * I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_MULT;
  --===========================================================================
  architecture ideal of IBIS_CCVS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Vout == Scale * Iin1 * Iin2;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source for taking the ratio
--              of two control input voltages. The output voltage value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) / V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  ZeroLimit   
--              DEFAULT ZeroLimit = '1.0e-3'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCVS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCVS_DIV;
  --===========================================================================
  architecture ideal of IBIS_VCVS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  quantity AbsVin2  : voltage;
  signal   NearZero : boolean := true;

  begin

    ---------------------------------------------------------------------------
    Selector : process (AbsVin2'above(ZeroLimit)) is
    begin
      if AbsVin2 > ZeroLimit then
        NearZero <= false;
      else
        NearZero <= true;
      end if;
    end process Selector;
    ---------------------------------------------------------------------------
    Message : process is
    begin
      wait on domain, NearZero;
      if AbsVin2 <= ZeroLimit then
        report "  The divisor in module 'IBIS_VCVS_DIV' is less than or equal to the " &
               "value defined in its parameter 'ZeroLimit = " & real'image(ZeroLimit) &
               "' at time = " & real'image(now) & "." severity WARNING;
      end if;
    end process Message;
    ---------------------------------------------------------------------------
    break on NearZero;
    ---------------------------------------------------------------------------

    AbsVin2 == abs(Vin2);
    if NearZero use
      Vout == Scale * Vin1 * Vin2 / ZeroLimit**2;
    else
      Vout == Scale * Vin1 / Vin2;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled voltage source for taking the ratio
--              of two control input currents. The output voltage value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) / I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  ZeroLimit   
--              DEFAULT ZeroLimit = '1.0e-3'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCVS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCVS_DIV;
  --===========================================================================
  architecture ideal of IBIS_CCVS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  quantity AbsIin2  : current;
  signal   NearZero : boolean := true;

  begin

    ---------------------------------------------------------------------------
    Selector : process (AbsIin2'above(ZeroLimit)) is
    begin
      if AbsIin2 > ZeroLimit then
        NearZero <= false;
      else
        NearZero <= true;
      end if;
    end process Selector;
    ---------------------------------------------------------------------------
    Message : process is
    begin
      wait on domain, NearZero;
      if AbsIin2 <= ZeroLimit then
        report "  The divisor in module 'IBIS_CCVS_DIV' is less than or equal to the " &
               "value defined in its parameter 'ZeroLimit = " & real'image(ZeroLimit) &
               "' at time = " & real'image(now) & "." severity WARNING;
      end if;
    end process Message;
    ---------------------------------------------------------------------------
    break on NearZero;
    ---------------------------------------------------------------------------

    AbsIin2 == abs(Iin2);
    Vin1    == 0.0;
    Vin2    == 0.0;
    if NearZero use
      Vout == Scale * Iin1 * Iin2 / ZeroLimit**2;
    else
      Vout == Scale * Iin1 / Iin2;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled voltage source with non-linear
--              relationship between control voltage and output voltage,
--              described using a piecewise-linear table.
--
--              Each voltage value in the array of X values is paired with
--              a Y array voltage value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control voltage determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output voltage. Linear interpolation is used for control
--              voltage values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control voltage)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output voltage)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VCVS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VCVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vout == Scale * Lookup(Vin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled voltage source with non-linear
--              relationship between control current and output voltage,
--              described using a piecewise-linear table.
--
--              Each current value in the array of X values is paired with
--              a Y array voltage value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control current determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output voltage. Linear interpolation is used for control
--              current values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control current)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output voltage)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CCVS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CCVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vin  == 0.0;
    Vout == Scale * Lookup(Iin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_TCVS_PWL (p, n)
--
-- DESCRIPTION:
--              A time-controlled voltage source with non-linear
--              relationship between time and output voltage, described
--              using a piecewise-linear table.
--
--              Each time value in the array of X values is paired with a
--              Y array votage value. The Length parameter must be set to
--              the number of elements in these arrays.
--
--              The arrays describe the output voltage waveform. The first
--              X value in the array gives the time at wich the waveform
--              begins in the simulation. Linear interpolation is used for
--              time values that do not fall exactly on a particular X
--              array value. Points outside the time range are
--              extrapolated horizontally, ie., the Y value of nearest
--              point is used.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (time)
--              DEFAULT X[1:Length] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  Y[1:Length]  Array of Y axis values (output voltage)
--              DEFAULT Y[1:Length] = '{0.00 0.10 0.90 1.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_TCVS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             Y_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n : electrical);

  end entity IBIS_TCVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_TCVS_PWL is

  quantity Vout across Iout through p to n;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vout == Scale * Lookup(now, X_vec, Y_vec, "HE");

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VECVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage event-controlled voltage source producing
--              time-based piecewise-linear voltage waveforms in response
--              to trigger events. The control voltage triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a voltage
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control voltage rises above Vth_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              voltage falls below Vth_F a falling edge waveform is
--              triggered, holding at the last voltage until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control voltage is above Vth_R,
--              and the first point of the rising edge waveform is output
--              if the control voltage is below Vth_F. If Vt_R is less
--              than or equal to Vth_F and the control voltage lies on or
--              between those values the high state voltage is output. If
--              Vth_R is greater than Vth_F and the control voltage lies
--              on or between those values, the average of the high and
--              low state voltages is output.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Vth_R       Threshold voltage to trigger a rising edge event
--              DEFAULT Vth_R = '0.0'
--
--  Vth_F       Threshold voltage to trigger a falling edge event
--              DEFAULT Vth_F = '0.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (voltage)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (voltage)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VECVS_PWL is

    generic (Vth_R : real := 0.0;
             Vth_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VECVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VECVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process (Vin'above(DC_threshold),
                                  Vin'above(Vth_R), Vin'above(Vth_F)) is
    ---------------------------------------------------------------------------
    begin

      if domain = quiescent_domain then     -- during DC analysis
        if Vin > DC_threshold then          -- high state
          EventState <= 1;
        else                                -- low state
          EventState <= -1;
        end if;
      else
        if Vin > Vth_R then                 -- rising edge ****************
          if EventState < 0 then            -- low to high state transition
            EventState <= 2;
          end if;
        elsif Vin < Vth_F then              -- falling edge ***************
          if EventState > 0 then            -- high to low state transition
            EventState <= -2;
          end if;
        end if;
      end if;

    end process Threshold_crossing;
    ---------------------------------------------------------------------------
    break on EventState;
    ---------------------------------------------------------------------------

    case EventState use
      when -2 =>        -- Executes after falling edge events (normal operation)
        Vout == Scale * Lookup(EventState'last_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes before first good threshold crossing
        Vout == Scale * YR_vec(YR_vec'left);        -- when input is LOW
      when 1 =>         -- Executes before first good threshold crossing
        Vout == Scale * YF_vec(YF_vec'left);        -- when input is HIGH
      when 2 =>         -- Executes after rising edge events (normal operation)
        Vout == Scale * Lookup(EventState'last_event, XR_vec, YR_vec, "HE");
      when others =>    -- This should really never execute
        Vout == 0.0;
    end case;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CECVS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current event-controlled voltage source producing
--              time-based piecewise-linear voltage waveforms in response
--              to trigger events. The control current triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a voltage
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control current rises above Ith_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              current falls below Ith_F a falling edge waveform is
--              triggered, holding at the last voltage until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control current is above Ith_R,
--              and the first point of the rising edge waveform is output
--              if the control current is below Ith_F. If It_R is less
--              than or equal to Ith_F and the control current lies on or
--              between those values the high state voltage is output. If
--              Ith_R is greater than Ith_F and the control current lies
--              on or between those values, the average of the high and
--              low state voltages is output.
--
-- TERMINALS:
--
--  p           positive voltage source terminal
--  n           negative voltage source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Ith_R       Threshold current to trigger a rising edge event
--              DEFAULT Ith_R = '0.0'
--
--  Ith_F       Threshold current to trigger a falling edge event
--              DEFAULT Ith_F = '0.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (current)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (current)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CECVS_PWL is

    generic (Ith_R : real := 0.0;
             Ith_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CECVS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CECVS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  constant  DC_threshold : real := (Ith_R + Ith_F) / 2.0;
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process (Iin'above(DC_threshold),
                                  Iin'above(Ith_R), Iin'above(Ith_F)) is
    ---------------------------------------------------------------------------
    begin

      if domain = quiescent_domain then     -- during DC analysis
        if Iin > DC_threshold then          -- high state
          EventState <= 1;
        else                                -- low state
          EventState <= -1;
        end if;
      else
        if Iin > Ith_R then                 -- rising edge ****************
          if EventState < 0 then            -- low to high state transition
            EventState <= 2;
          end if;
        elsif Iin < Ith_F then              -- falling edge ***************
          if EventState > 0 then            -- high to low state transition
            EventState <= -2;
          end if;
        end if;
      end if;

    end process Threshold_crossing;
    ---------------------------------------------------------------------------
    break on EventState;
    ---------------------------------------------------------------------------

    Vin == 0.0;
    case EventState use
      when -2 =>        -- Executes after falling edge events (normal operation)
        Vout == Scale * Lookup(EventState'last_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes before first good threshold crossing
        Vout == Scale * YR_vec(YR_vec'left);        -- when input is LOW
      when 1 =>         -- Executes before first good threshold crossing
        Vout == Scale * YF_vec(YF_vec'left);        -- when input is HIGH
      when 2 =>         -- Executes after rising edge events (normal operation)
        Vout == Scale * Lookup(EventState'last_event, XR_vec, YR_vec, "HE");
      when others =>    -- This should really never execute
        Vout == 0.0;
    end case;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_I (p, n)
--
-- DESCRIPTION:
--              A DC current source. Output current is calculated:
--
--              I(p,n) = Idc * Scale
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--
-- PARAMETERS:
--
--  Idc         DC current value
--              DEFAULT Idc = '1.0'
--
--  Scale       Scaling factor for calculating output voltage
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 pada vss IBIS_I Idc=0.025
--
--  VAMS:       IBIS_I #(.Idc(0.025)) I1 (pada, vss);
--
--  VHDL:
--        entity IBIS_I is
--          port (terminal pada, vss : electrical);
--        end entity IBIS_I;
--        architecture IBIS_SIM of IBIS_I is
--        begin
--          generic map ( Idc => 0.025 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_I is

    generic (Idc   : real := 1.0;
             Scale : real := 1.0);

    port (terminal p, n : electrical);

  end entity IBIS_I;
  --===========================================================================
  architecture ideal of IBIS_I is

  quantity Vout across Iout through p to n;

  begin

    Iout == Scale * Idc;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source. The output current
--              value is calculated:
--
--              I(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 pupwr pupwra vcp vss IBIS_VCCS Scale=0.05
--
--  VAMS:       IBIS_VCCS #(.Scale(0.05)) VCCS1 (pupwr, pupwr, vcp, vss);
--
--  VHDL:
--        entity IBIS_VCCS is
--          port (terminal pupwr, pupwr, vcp, vss : electrical);
--        end entity IBIS_VCCS;
--        architecture IBIS_SIM of IBIS_VCCS is
--        begin
--          generic map ( Scale => 0.05 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS;
  --===========================================================================
  architecture ideal of IBIS_VCCS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * Vin;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source. The output current
--              value is calculated:
--
--              I(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS Scale=0.1
--
--  VAMS:       IBIS_CCVS #(.Scale(0.1)) CCVS1 (buf2in, vss, pad, padi);
--
--  VHDL:
--        entity IBIS_CCVS is
--          port (terminal buf2in, vss, pad, padi : electrical);
--        end entity IBIS_CCVS;
--        architecture IBIS_SIM of IBIS_CCVS is
--        begin
--          generic map ( Scale => 0.1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS;
  --===========================================================================
  architecture ideal of IBIS_CCCS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * Iin;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source with delay. All
--              changes at the sense terminals affect the output later in
--              time, using the specified delay. The output current value
--              is calculated:
--
--              I(p,n) = V(ps,ns) * Scale, delayed by TD
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  TD          Time delay from control voltage change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_VCCS_DELAY TD=200p
--
--  VAMS:       IBIS_VCCS_DELAY #(.TD(200e-12)) \
--              VCCS1 (buf2in, vss, pad, padi);
--
--  VHDL:
--        entity IBIS_VCCS_DELAY is
--          port (terminal buf2in, vss, pad, padi : electrical);
--        end entity IBIS_VCCS_DELAY;
--        architecture IBIS_SIM of IBIS_VCCS_DELAY is
--        begin
--          generic map ( TD => 200e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_VCCS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * Vin'delayed(TD);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_DELAY (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source with delay. All
--              changes at the sense terminals affect the output later,
--              using the specified delay in seconds. The current value is
--              calculated:
--
--              I(p,n) = I(ps,ns) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  TD          Time delay from control current change to output change
--              DEFAULT TD = '0.0'
--
--  Scale       Scaling factor for calculating current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 buf2in vss pad padi IBIS_CCVS_DELAY Scale=100 TD=200p
--
--  VAMS:       IBIS_CCVS_DELAY #(.Scale(100),.TD(200e-12)) \
--              CCVS1 (buf2in, vss, pad, padi);
--
--  VHDL:
--        entity IBIS_CCVS_DELAY is
--          port (terminal buf2in, vss, pad, padi : electrical);
--        end entity IBIS_CCVS_DELAY;
--        architecture IBIS_SIM of IBIS_CCVS_DELAY is
--        begin
--          generic map ( Scale => 100, TD => 200e-12 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_DELAY is

    generic (TD    : real := 0.0;
             Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_DELAY;
  --===========================================================================
  architecture ideal of IBIS_CCCS_DELAY is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * Iin'delayed(TD);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the lesser
--              of two control input voltages. The output current value is
--              calculated:
--
--              I(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCCS_MIN
--
--  VAMS:       IBIS_VCCS_MIN #(.Scale(1)) MIN1
--              (bufctl, vss, ctlmain, vss, boost, vss);
--
--  VHDL:
--        entity IBIS_VCCS_MIN is
--          port (terminal bufctl, vss, ctlmain, vss,
--          boost, vss : electrical);
--        end entity IBIS_VCCS_MIN;
--        architecture IBIS_SIM of IBIS_VCCS_MIN is
--        begin
--          generic map ( Scale => 1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCCS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MIN;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * realmin(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_MIN (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the lesser
--              of two control input voltages. The output current value is
--              calculated:
--
--              I(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain ctlmaini boost boosti IBIS_CCVS_MIN
--
--  VAMS:       IBIS_CCVS_MIN #(.Scale(1))
--              MIN1 (bufctl, vss, ctlmain, vss, boost, vss);
--
--  VHDL:
--        entity IBIS_CCVS_MIN is
--          port (terminal bufctl, vss, ctlmain, vss,
--          boost, vss : electrical);
--        end entity IBIS_CCVS_MIN;
--        architecture IBIS_SIM of IBIS_CCVS_MIN is
--        begin
--          generic map ( Scale => 1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCCS_MIN is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MIN;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MIN is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * realmin(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the greater
--              of two control input voltages. The output current value is
--              calculated:
--
--              I(p,n) = min( V(ps1,ns1) V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--  SPICE:      X1 bufctl vss ctlmain vss boost vss IBIS_VCCS_MAX
--
--  VAMS:       IBIS_VCCS_MAX #(.Scale(1)) MAX1
--              (bufctl, vss, ctlmain, vss, boost, vss);
--
--  VHDL:
--        entity IBIS_VCCS_MAX is
--          port (terminal bufctl, vss, ctlmain, vss,
--          boost, vss : electrical);
--        end entity IBIS_VCCS_MAX;
--        architecture IBIS_SIM of IBIS_VCCS_MAX is
--        begin
--          generic map ( Scale => 1 )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_VCCS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MAX;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * realmax(Vin1, Vin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_MAX (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the greater
--              of two input currents. The output current value is
--              calculated:
--
--              V(p,n) = max( I(ps1,ns1) I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;
--=============================================================================
  entity IBIS_CCCS_MAX is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MAX;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MAX is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * realmax(Iin1, Iin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the
--              absolute value of the control voltage. The output current
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( V(ps,ns) ) * Scale
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_ABS;
  --===========================================================================
  architecture ideal of IBIS_VCCS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;

  begin

    Iout == Scale * abs(Vin);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_ABS (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the
--              absolute value of the control current. The output current
--              value is always positive, and is calculated:
--
--              V(p,n) = abs( I(ps,ns) ) * Scale
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive resistor terminal
--  n           negative resistor terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_ABS is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_ABS;
  --===========================================================================
  architecture ideal of IBIS_CCCS_ABS is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;

  begin

    Vin  == 0.0;
    Iout == Scale * abs(Iin);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the sum of
--              two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) + V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_SUM;
  --===========================================================================
  architecture ideal of IBIS_VCCS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * (Vin1 + Vin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_SUM (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the sum of
--              two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) + I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_SUM is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_SUM;
  --===========================================================================
  architecture ideal of IBIS_CCCS_SUM is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * (Iin1 + Iin2);

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the product
--              of two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) * V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_MULT;
  --===========================================================================
  architecture ideal of IBIS_VCCS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  begin

    Iout == Scale * Vin1 * Vin2;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_MULT (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the product
--              of two control input currents. The output current value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) * I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_MULT is

    generic (Scale : real := 1.0);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_MULT;
  --===========================================================================
  architecture ideal of IBIS_CCCS_MULT is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  begin

    Vin1 == 0.0;
    Vin2 == 0.0;
    Iout == Scale * Iin1 * Iin2;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A voltage-controlled current source for taking the ratio
--              of two control input voltages. The output current value is
--              calculated:
--
--              V(p,n) = ( V(ps1,ns1) / V(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 draw no current. The sense
--              terminals ps2 and ns2 draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control voltage 1
--  ns1         negative sense terminal for control voltage 1
--  ps2         positive sense terminal for control voltage 2
--  ns2         negative sense terminal for control voltage 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  ZeroLimit   
--              DEFAULT ZeroLimit = '1.0e-3'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_VCCS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_VCCS_DIV;
  --===========================================================================
  architecture ideal of IBIS_VCCS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across              ps1 to ns1;
  quantity Vin2 across              ps2 to ns2;

  quantity AbsVin2  : voltage;
  signal   NearZero : boolean := true;

  begin

    ---------------------------------------------------------------------------
    Selector : process (AbsVin2'above(ZeroLimit)) is
    begin
      if AbsVin2 > ZeroLimit then
        NearZero <= false;
      else
        NearZero <= true;
      end if;
    end process Selector;
    ---------------------------------------------------------------------------
    Message : process is
    begin
      wait on domain, NearZero;
      if AbsVin2 <= ZeroLimit then
        report "  The divisor in module 'IBIS_VCCS_DIV' is less than or equal to the " &
               "value defined in its parameter 'ZeroLimit = " & real'image(ZeroLimit) &
               "' at time = " & real'image(now) & "." severity WARNING;
      end if;
    end process Message;
    ---------------------------------------------------------------------------
    break on NearZero;
    ---------------------------------------------------------------------------

    AbsVin2 == abs(Vin2);
    if NearZero use
      Iout == Scale * Vin1 * Vin2 / ZeroLimit**2;
    else
      Iout == Scale * Vin1 / Vin2;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_CCCS_DIV (p, n, ps1, ns1, ps2, ns2)
--
-- DESCRIPTION:
--              A current-controlled current source for taking the ratio
--              of two control input currents. The output current value is
--              calculated:
--
--              V(p,n) = ( I(ps1,ns1) / I(ps2,ns2) ) * Scale
--
--              The sense terminals ps1 and ns1 behave as a short circuit.
--              The sense terminals ps2 and ns2 behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps1         positive sense terminal for control current 1
--  ns1         negative sense terminal for control current 1
--  ps2         positive sense terminal for control current 2
--  ns2         negative sense terminal for control current 2
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  ZeroLimit   
--              DEFAULT ZeroLimit = '1.0e-3'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_CCCS_DIV is

    generic (Scale     : real := 1.0;
             ZeroLimit : real := 1.0e-3);

    port (terminal p, n, ps1, ns1, ps2, ns2 : electrical);

  end entity IBIS_CCCS_DIV;
  --===========================================================================
  architecture ideal of IBIS_CCCS_DIV is

  quantity Vout across Iout through p   to n;
  quantity Vin1 across Iin1 through ps1 to ns1;
  quantity Vin2 across Iin2 through ps2 to ns2;

  quantity AbsIin2  : current;
  signal   NearZero : boolean := true;

  begin

    ---------------------------------------------------------------------------
    Selector : process (AbsIin2'above(ZeroLimit)) is
    begin
      if AbsIin2 > ZeroLimit then
        NearZero <= false;
      else
        NearZero <= true;
      end if;
    end process Selector;
    ---------------------------------------------------------------------------
    Message : process is
    begin
      wait on domain, NearZero;
      if AbsIin2 <= ZeroLimit then
        report "  The divisor in module 'IBIS_CCCS_DIV' is less than or equal to the " &
               "value defined in its parameter 'ZeroLimit = " & real'image(ZeroLimit) &
               "' at time = " & real'image(now) & "." severity WARNING;
      end if;
    end process Message;
    ---------------------------------------------------------------------------
    break on NearZero;
    ---------------------------------------------------------------------------

    AbsIin2 == abs(Iin2);
    Vin1    == 0.0;
    Vin2    == 0.0;
    if NearZero use
      Iout == Scale * Iin1 * Iin2 / ZeroLimit**2;
    else
      Iout == Scale * Iin1 / Iin2;
    end use;

  end architecture ideal;
--=============================================================================
--
-- MODULE:      IBIS_VCCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage-controlled current source with non-linear
--              relationship between control voltage and output current,
--              described using a piecewise-linear table.
--
--              Each voltage value in the array of X values is paired with
--              a Y array current value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control voltage determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output current. Linear interpolation is used for control
--              voltage values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns draw no current.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control voltage)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output current)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VCCS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VCCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Iout == Scale * Lookup(Vin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_CCCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current-controlled current source with non-linear
--              relationship between control current and output current,
--              described using a piecewise-linear table.
--
--              Each current value in the array of X values is paired with
--              a Y array current value. The Length parameter must be set
--              to the number of elements in these arrays.
--
--              The input control current determines a position in the X
--              array, and the corersponding Y array value is used for the
--              output current. Linear interpolation is used for control
--              current values that do not fall exactly on a particular X
--              array value. Points outside the X array range are
--              extrapolated using the slope of the nearest two points of
--              the PWL data.
--
--              The sense terminals ps and ns behave as a short circuit.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (input control current)
--              DEFAULT X[1:Length] = '{-5.00 0.00 5.00 10.00}'
--
--  Y[1:Length]  Array of Y axis values (output current)
--              DEFAULT Y[1:Length] = '{-0.10 0.00 0.10 0.10}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CCCS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
             Y_data : real_vector := (-0.10,  0.00,  0.10,  0.10);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CCCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Vin  == 0.0;
    Iout == Scale * Lookup(Iin, X_vec, Y_vec, "SE");

  end architecture ideal;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_TCCS_PWL (p, n)
--
-- DESCRIPTION:
--              A time-controlled current source with non-linear
--              relationship between time and output current, described
--              using a piecewise-linear table.
--
--              Each time value in the array of X values is paired with a
--              Y array votage value. The Length parameter must be set to
--              the number of elements in these arrays.
--
--              The arrays describe the output current waveform. The first
--              X value in the array gives the time at wich the waveform
--              begins in the simulation. Linear interpolation is used for
--              time values that do not fall exactly on a particular X
--              array value. Points outside the time range are
--              extrapolated horizontally, ie., the Y value of nearest
--              point is used.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--
-- PARAMETERS:
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length      Number of elements in the X and Y arrays
--              DEFAULT Length = '4'
--
--  X[1:Length]  Array of X axis values (time)
--              DEFAULT X[1:Length] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  Y[1:Length]  Array of Y axis values (output current)
--              DEFAULT Y[1:Length] = '{0.00 0.10 0.90 1.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_TCCS_PWL is

    generic (Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             X_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             Y_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n : electrical);

  end entity IBIS_TCCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_TCCS_PWL is

  quantity Vout across Iout through p to n;
  --===========================================================================
  constant  X_vec : real_vector := FileRead(DataFile, "X_data", X_data, Max_length);
  constant  Y_vec : real_vector := FileRead(DataFile, "Y_data", Y_data, Max_length);
--=============================================================================
  begin

    Iout == Scale * Lookup(now, X_vec, Y_vec, "HE");

  end architecture ideal;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_VECCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A voltage event-controlled current source producing
--              time-based piecewise-linear current waveforms in response
--              to trigger events. The control voltage triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a current
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control voltage rises above Vth_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              voltage falls below Vth_F a falling edge waveform is
--              triggered, holding at the last current until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control voltage is above Vth_R,
--              and the first point of the rising edge waveform is output
--              if the control voltage is below Vth_F. If Vt_R is less
--              than or equal to Vth_F and the control voltage lies on or
--              between those values the high state current is output. If
--              Vth_R is greater than Vth_F and the control voltage lies
--              on or between those values, the average of the high and
--              low state currents is output.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control voltage
--  ns          negative sense terminal for control voltage
--
-- PARAMETERS:
--
--  Vth_R       Threshold voltage to trigger a rising edge event
--              DEFAULT Vth_R = '0.0'
--
--  Vth_F       Threshold voltage to trigger a falling edge event
--              DEFAULT Vth_F = '0.0'
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (current)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (current)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_VECCS_PWL is

    generic (Vth_R : real := 0.0;
             Vth_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------
    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_VECCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_VECCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across              ps to ns;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process (Vin'above(DC_threshold),
                                  Vin'above(Vth_R), Vin'above(Vth_F)) is
    ---------------------------------------------------------------------------
    begin

      if domain = quiescent_domain then     -- during DC analysis
        if Vin > DC_threshold then          -- high state
          EventState <= 1;
        else                                -- low state
          EventState <= -1;
        end if;
      else
        if Vin > Vth_R then                 -- rising edge ****************
          if EventState < 0 then            -- low to high state transition
            EventState <= 2;
          end if;
        elsif Vin < Vth_F then              -- falling edge ***************
          if EventState > 0 then            -- high to low state transition
            EventState <= -2;
          end if;
        end if;
      end if;

    end process Threshold_crossing;
    ---------------------------------------------------------------------------
    break on EventState;
    ---------------------------------------------------------------------------

    case EventState use
      when -2 =>        -- Executes after falling edge events (normal operation)
        Iout == Scale * Lookup(EventState'last_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes before first good threshold crossing
        Iout == Scale * YR_vec(YR_vec'left);        -- when input is LOW
      when 1 =>         -- Executes before first good threshold crossing
        Iout == Scale * YF_vec(YF_vec'left);        -- when input is HIGH
      when 2 =>         -- Executes after rising edge events (normal operation)
        Iout == Scale * Lookup(EventState'last_event, XR_vec, YR_vec, "HE");
      when others =>    -- This should really never execute
        Iout == 0.0;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_CECCS_PWL (p, n, ps, ns)
--
-- DESCRIPTION:
--              A current event-controlled current source producing
--              time-based piecewise-linear current waveforms in response
--              to trigger events. The control current triggers separate
--              rising and falling PWL waveforms to effect transitions
--              between high and low states.
--
--              Each time value in the XR array is paired with a current
--              value in the YR array to describe a rising edge waveform.
--              Likewise for XF and YF arrays, which describe a falling
--              edge waveform. The Length_R and Length_F parameters must
--              be set to the number of elements in the *R and *F arrays,
--              respectively.
--
--              When the control current rises above Ith_R the rising edge
--              waveform is output. Voltage then remains at the level of
--              the last point in the rising waveform table until a
--              falling edge transition is triggerd. When the control
--              current falls below Ith_F a falling edge waveform is
--              triggered, holding at the last current until the next
--              rising edge.
--
--              In DC analysis the first point of the falling edge
--              waveform is output if the control current is above Ith_R,
--              and the first point of the rising edge waveform is output
--              if the control current is below Ith_F. If It_R is less
--              than or equal to Ith_F and the control current lies on or
--              between those values the high state current is output. If
--              Ith_R is greater than Ith_F and the control current lies
--              on or between those values, the average of the high and
--              low state currents is output.
--
-- TERMINALS:
--
--  p           positive current source terminal
--  n           negative current source terminal
--  ps          positive sense terminal for control current
--  ns          negative sense terminal for control current
--
-- PARAMETERS:
--
--  Ith_R       Threshold current to trigger a rising edge event
--              DEFAULT Ith_R = '0.0'
--
--  Ith_F       Threshold current to trigger a falling edge event
--              DEFAULT Ith_F = '0.0'
--
--  Scale       Scaling factor for calculating output current
--              DEFAULT Scale = '1.0'
--
--  Length_R    Number of elements in the XR and YR arrays
--              DEFAULT Length_R = '4'
--
--  Length_F    Number of elements in the XF and YF arrays
--              DEFAULT Length_F = '4'
--
--  XR[1:Length_R]  Array of rising edge X values (time)
--              DEFAULT XR[1:Length_R] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YR[1:Length_R]  Array of rising edge Y values (current)
--              DEFAULT YR[1:Length_R] = '{0.00 0.10 0.90 1.00}'
--
--  XF[1:Length_F]  Array of falling edge X values (time)
--              DEFAULT XF[1:Length_F] = '{0.50e-9 1.00e-9 2.00e-9 2.50e-9}'
--
--  YF[1:Length_F]  Array of falling edge Y values (current)
--              DEFAULT YF[1:Length_F] = '{1.00 0.90 0.10 0.00}'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
  entity IBIS_CECCS_PWL is

    generic (Ith_R : real := 0.0;
             Ith_F : real := 0.0;
             Scale : real := 1.0;
             ------------------------------------------------------------------
             -- Vectors of the PWL table
             ------------------------------------------------------------------
             XR_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YR_data : real_vector := (0.00,     0.10,     0.90,     1.00);
             XF_data : real_vector := (0.50e-9,  1.00e-9,  2.00e-9,  2.50e-9);
             YF_data : real_vector := (1.00,     0.90,     0.10,     0.00);
             ------------------------------------------------------------------
             DataFile   : string  := "";      -- Data file name
             Max_length : integer := 100);    -- Maximum possible table length
    ---------------------------------------------------------------------------

    port (terminal p, n, ps, ns : electrical);

  end entity IBIS_CECCS_PWL;
  --===========================================================================
  architecture ideal of IBIS_CECCS_PWL is

  quantity Vout across Iout through p  to n;
  quantity Vin  across Iin  through ps to ns;
  signal   EventState : integer := 0;
  --===========================================================================
  constant  XR_vec : real_vector := FileRead(DataFile, "XR_data", XR_data, Max_length);
  constant  YR_vec : real_vector := FileRead(DataFile, "YR_data", YR_data, Max_length);
  constant  XF_vec : real_vector := FileRead(DataFile, "XF_data", XF_data, Max_length);
  constant  YF_vec : real_vector := FileRead(DataFile, "YF_data", YF_data, Max_length);
--=============================================================================
  constant  DC_threshold : real := (Ith_R + Ith_F) / 2.0;
--=============================================================================
  begin

    ---------------------------------------------------------------------------
    Threshold_crossing : process (Iin'above(DC_threshold),
                                  Iin'above(Ith_R), Iin'above(Ith_F)) is
    ---------------------------------------------------------------------------
    begin

      if domain = quiescent_domain then     -- during DC analysis
        if Iin > DC_threshold then          -- high state
          EventState <= 1;
        else                                -- low state
          EventState <= -1;
        end if;
      else
        if Iin > Ith_R then                 -- rising edge ****************
          if EventState < 0 then            -- low to high state transition
            EventState <= 2;
          end if;
        elsif Iin < Ith_F then              -- falling edge ***************
          if EventState > 0 then            -- high to low state transition
            EventState <= -2;
          end if;
        end if;
      end if;

    end process Threshold_crossing;
    ---------------------------------------------------------------------------
    break on EventState;
    ---------------------------------------------------------------------------

    Vin == 0.0;
    case EventState use
      when -2 =>        -- Executes after falling edge events (normal operation)
        Iout == Scale * Lookup(EventState'last_event, XF_vec, YF_vec, "HE");
      when -1 =>        -- Executes before first good threshold crossing
        Iout == Scale * YR_vec(YR_vec'left);        -- when input is LOW
      when 1 =>         -- Executes before first good threshold crossing
        Iout == Scale * YF_vec(YF_vec'left);        -- when input is HIGH
      when 2 =>         -- Executes after rising edge events (normal operation)
        Iout == Scale * Lookup(EventState'last_event, XR_vec, YR_vec, "HE");
      when others =>    -- This should really never execute
        Iout == 0.0;
    end case;

  end architecture ideal;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_T (n1, ref1, n2, ref2)
--
-- DESCRIPTION:
--              A lossless transmission line.
--
-- TERMINALS:
--
--  n1          Signal terminal for end 1
--  ref1        Reference terminal for end 1
--  n2          Signal terminal for end 2
--  ref2        Reference terminal for end 2
--
-- PARAMETERS:
--
--  Z0          Transmission line impedance in ohms
--              DEFAULT Z0 = '50.0'
--
--  TD          Time delay in seconds
--              DEFAULT TD = '1.0e-9'
--
--
-- INSTANCE CALL EXAMPLES:
--
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
--=============================================================================
  entity IBIS_T is

    generic (Z0 : real := 50.0;
             TD : real := 1.0e-9);

    port (terminal n1, ref1, n2, ref2 : electrical);

  end entity IBIS_T;
  --===========================================================================
  architecture Ideal of IBIS_T is

  quantity Vend1 across Iend1 through n1 to ref1;
  quantity Vend2 across Iend2 through n2 to ref2;

  quantity i12 : current;    -- Current wave traveling from end1 to end2
  quantity i21 : current;    -- Current wave traveling from end2 to end1

  begin

    Iend1 == i12 - i21'delayed(TD);
    Iend2 == i21 - i12'delayed(TD);
    Vend1 == Z0 * (2.0 * i12 - Iend1);
    Vend2 == Z0 * (2.0 * i21 - Iend2);

  end architecture Ideal;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_INPUT (PC_ref, GC_ref, Pad, Rcv_D)
--
-- DESCRIPTION:
--              An IBIS Input buffer. The effects modeled are:
--              - Clamp I/V curves
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--              - Threshold based received logic detection.
--
--              The voltage at the Pad terminal is monitored and
--              a corresponding output voltage is produced at the Rcv_D
--              terminal. The Vinl and Vinh parameters are used as voltage
--              thresholds for controlling the core-level output data
--              at the Rcv_D terminal. When the Pad voltage is above
--              Vinh the Rcv_D voltage is 1.0. When the Pad voltage is
--              below Vinl the Rcv_D voltage is 0.0. When the Pad voltage
--              is in the middle region the Rcv_D voltage is 0.5.
--
--              Clamp currents can be experimentally scaled using the
--              kI_pc and kI_gc parameters. These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--
-- TERMINALS:
--
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  Rcv_D       Received data from buffer to core logic
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  kI_pc       Scaling factor for power clamp current values.
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT kI_pc = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT kI_gc = '1.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_INPUT  #(
--   .C_comp(1.2e-12), .Vinh(2.0), .Vinl(0.8),
--   .Vpc_ref(5.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}))
--  B1  (B1_pc, B1_gc, B1_pad, B1_a);
--
--  VHDL:
--        entity IBIS_INPUT is
--          port (terminal B1_pc, B1_gc, B1_pad, B1_a : electrical);
--        end entity IBIS_INPUT;
--        architecture IBIS_SIM of IBIS_INPUT is
--        begin
--          generic map ( C_comp => 1.2e-12, Vinh => 2.0,
--          Vinl => 0.8, Vpc_ref => 5.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_INPUT is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.50;       -- splitting coefficients
           kC_comp_gc : real := 0.50;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           MaxIV_length : integer := 100);    -- Maximum possible length of IV tables
     --------------------------------------------------------------------------
     port (terminal PC_ref : electrical;
           terminal GC_ref : electrical;

           terminal Pad    : electrical;
           terminal Rcv_D  : electrical);

end entity IBIS_INPUT;
--=============================================================================
architecture SIMPLE_RECEIVER of IBIS_INPUT is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;
  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);
--=============================================================================
begin
  --===========================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc*Lookup(Vpc, Vpc_vec, Ipc_vec, "SE") + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Igc ==      kI_gc*Lookup(Vgc, Vgc_vec, Igc_vec, "SE") + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
  if        Vgc'above(Vinh_val)  use Vrcv == 1.0;
  elsif not(Vgc'above(Vinl_val)) use Vrcv == 0.0;
  else                               Vrcv == 0.5;
  end use;
  break on Vgc'above(Vinh_val), Vgc'above(Vinl_val);
--=============================================================================
end architecture SIMPLE_RECEIVER;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_OUTPUT (PU_ref, PD_ref, Pad, In_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS Output buffer. The effects modeled are:
--              - Pullup, pulldown, and clamp I/V curves with IBIS 2.1
--                waveform-based pullup/pulldown turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--
--              The Vth_R and Vth_F parameters control how the In_D 
--              terminal behaves. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D terminal is
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--
--              The Pad terminal is driven by either the
--              Pullup or Pulldown I/V table, depending on the low/high
--              state of the In_D terminal. State changes at the In_D
--              terminal trigger controlled transitions to the other
--              I/V table, using the waveform and test fixture parameters
--              to guide the transition.
--
--              Clamp, Pullup, and Pulldown currents can be experimentally
--              scaled using the kI_pc, kI_gc, kI_pu, and kI_pd parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpu_length Number of elements in the Ipu_data and Vpu_data arrays
--              Count the number of IBIS [Pullup] entries to get
--              this value.
--              DEFAULT IVpu_length = '4'
--
--  IVpd_length Number of elements in the Ipd_data and Vpd_data arrays
--              Count the number of IBIS [Pulldown] entries to get
--              this value.
--              DEFAULT IVpd_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipu_data[1:IVpu_length]  Array of current values for pullup I/V
--              Extract the IBIS [Pullup] current column to get this.
--              DEFAULT Ipu_data[1:IVpu_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vpu_data[1:IVpu_length]  Array of voltage values for pullup I/V
--              Extract the IBIS [Pullup] voltage column to get this.
--              DEFAULT Vpu_data[1:IVpu_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipd_data[1:IVpd_length]  Array of current values for pulldown I/V
--              Extract the IBIS [Pulldown] current column to get this.
--              DEFAULT Ipd_data[1:IVpd_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpd_data[1:IVpd_length]  Array of voltage values for pulldown I/V
--              Extract the IBIS [Pulldown] voltage column to get this.
--              DEFAULT Vpd_data[1:IVpd_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with V_fixture = 0 to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTr2_length  Number of elements in the Tr2 and Vr2 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTr2_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  VTf2_length  Number of elements in the Tf2 and Vf2 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT VTf2_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vr2_data[1:VTr2_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr2_data[1:VTr2_length] = '{2.50 2.50 5.00 5.00}'
--
--  Tr2_data[1:VTr2_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Tr2_data[1:VTr2_length] = '{0.00 0.50e-9 0.80e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf2_data[1:VTf2_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the higher the higher V_fixture to get this value.
--              DEFAULT Vf2_data[1:VTf2_length] = '{2.50 2.50 0.00 0.00}'
--
--  Tf2_data[1:VTf2_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the higher the higher V_fixture to get this value.
--              DEFAULT Tf2_data[1:VTf2_length] = '{0.00 0.50e-9 0.80e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_r2      V_fixture value for Rising Waveform 2
--              Extract the higher V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r2 = '5.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Vfx_f2      V_fixture value for Falling Waveform 2
--              Extract the lower V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f2 = '0.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with lower V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_r2      R_fixture value for Rising Waveform 2
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with higher V_fixture.
--              DEFAULT Rfx_r2 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with higher V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  Rfx_f2      R_fixture value for Falling Waveform2
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with lower V_fixture.
--              DEFAULT Rfx_f2 = '50.0 from (0:inf)'
--
--  kI_pc       Scaling factor for power clamp current values.
--              DEFAULT kI_pc = '1.0'
--
--  kI_pu       Scaling factor for pullup clamp current values.
--              DEFAULT kI_pu = '1.0'
--
--  kI_pd       Scaling factor for pulldown current values.
--              DEFAULT kI_pd = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              DEFAULT kI_gc = '1.0'
--
--  kt_rise     Scaling factor for [Rising waveform] time values.
--              DEFAULT kt_rise = '1.0 from (0:inf)'
--
--  kt_fall     Scaling factor for [Falling waveform] time values.
--              DEFAULT kt_fall = '1.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum dt between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_OUTPUT  #(
--   .C_comp(1.2e-12),
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpu_length(4), .I_pu({ 0.10,  0.00, -0.10, -0.20}),
--                    .V_pu({-5.00,  0.00,  5.00, 10.00}),
--   .IVpd_length(4), .I_pd({-0.10,  0.00,  0.10,  0.20}),
--                    .V_pd({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTr2_length(6), .Vr2({1.66, 1.66,    2.50,    2.50,    3.33, 3.33}),
--                    .Tr2({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.0e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf2_length(4), .Vf2({1.66, 1.66,    0.00,    0.00}),
--                    .Tf2({0.00, 0.50e-9, 0.80e-9, 3.00e-9}) )
--  B1  (B1_pu,  B1_pd,  B1_pad, B1_in,  B1_pc,    B1_gc);
--
--  VHDL:
--        entity IBIS_OUTPUT is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_pc, B1_gc : electrical);
--        end entity IBIS_OUTPUT;
--        architecture IBIS_SIM of IBIS_OUTPUT is
--        begin
--          generic map ( C_comp => 1.2e-12, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_OUTPUT is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vr2_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf2_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_r2 : real := 5.0;
           Vfx_f1 : real := 5.0;
           Vfx_f2 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_r2 : real := 50.0;
           Rfx_f1 : real := 50.0;
           Rfx_f2 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_OUTPUT;
--=============================================================================
architecture IBIS_2EQ2UK of IBIS_OUTPUT is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;

  signal    PU_State      : integer := 0;
  signal    PD_State      : integer := 0;
  quantity  kpu           : current := 0.0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vr2_vec     : real_vector := FileRead(DataFile, "Vr2_data", Vr2_data, MaxVt_length);
  constant  Tr2_vec     : real_vector := FileRead(DataFile, "Tr2_data", Tr2_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);
  constant  Vf2_vec     : real_vector := FileRead(DataFile, "Vf2_data", Vf2_data, MaxVt_length);
  constant  Tf2_vec     : real_vector := FileRead(DataFile, "Tf2_data", Tf2_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_r2_val  : real := FileRead(DataFile, "Vfx_r2", Vfx_r2);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);
  constant  Vfx_f2_val  : real := FileRead(DataFile, "Vfx_f2", Vfx_f2);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_r2_val  : real := FileRead(DataFile, "Rfx_r2", Rfx_r2);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
  constant  Rfx_f2_val  : real := FileRead(DataFile, "Rfx_f2", Rfx_f2);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tr2_vec, Tf1_vec, Tf2_vec);

  type k_elements    is (Tcm, Kr1, Kr2, Kf1, Kf2);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_r2   : real_vector;
                  T_r2   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;
                  V_f2   : real_vector;
                  T_f2   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_r2  : real;
                  V_fx_f1  : real;
                  V_fx_f2  : real;

                  R_fx_r1  : real;
                  R_fx_r2  : real;
                  R_fx_f1  : real;
                  R_fx_f2  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vr2cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf2cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_r2 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;
    variable dVwfm_f2 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Iout2 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipc2  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipu2  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Ipd2  : real := 0.0;
    variable Igc1  : real := 0.0;
    variable Igc2  : real := 0.0;

    variable Term1 : real := 0.0;
    variable Term2 : real := 0.0;
    variable Den   : real := 1.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_r2, T_f1, T_f2);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vr2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r2, V_r2, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");
    Vf2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f2, V_f2, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vr2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r2, V_r2, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
        Vf2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f2, V_f2, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " R2: " & real'image(Vr2cm(2)) &
      --       " F1: " & real'image(Vf1cm(2)) &
      --       " F2: " & real'image(Vf2cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_r2 := 0.0;
        dVwfm_f1 := 0.0;
        dVwfm_f2 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_r2 := ((Vr2cm(2) - Vr2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr2cm(3) - Vr2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f2 := ((Vf2cm(2) - Vf2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf2cm(3) - Vf2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dR2: " & real'image(dVwfm_r2) &
      --       " dF1: " & real'image(dVwfm_f1) &
      --       " dF2: " & real'image(dVwfm_f2);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 Kr2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;
      Iout2 := ((Vr2cm(2) - V_fx_r2) / R_fx_r2) + C_comp_total * dVwfm_r2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vr2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vr2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vr2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vr2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on and Kr2 == Kpd_off
      K_tables(Kr1)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      K_tables(Kr2)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 Kf2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;
      Iout2 := ((Vf2cm(2) - V_fx_f2) / R_fx_f2) + C_comp_total * dVwfm_f2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vf2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vf2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vf2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vf2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on and Kf2 == Kpu_off
      K_tables(Kf1)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      K_tables(Kf2)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kr2: " & real'image(K_tables(Kr2)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i)) &
      --       " Kf2: " & real'image(K_tables(Kf2)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vr2cm(1) := Vr2cm(2);
      Vf1cm(1) := Vf1cm(2);
      Vf2cm(1) := Vf2cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vr2cm(2) := Vr2cm(3);
      Vf1cm(2) := Vf1cm(3);
      Vf2cm(2) := Vf2cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vr2_vec, Tr2_vec,
                                            Vf1_vec, Tf1_vec,
                                            Vf2_vec, Tf2_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_r2_val,
                                            Vfx_f1_val,
                                            Vfx_f2_val,

                                            Rfx_r1_val,
                                            Rfx_r2_val,
                                            Rfx_f1_val,
                                            Rfx_f2_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Vin > DC_threshold) then                              -- high state
        PU_State <=  1;
        PD_State <= -1;
      else                                                      -- low state
        PU_State <= -1;
        PD_State <=  1;
      end if;
    else
      if (PU_State > 0) and (PD_State < 0) then                 -- curent state is high
        if (Vin < Vth_F) then                                   -- go to low state
          PU_State <= -2;
          PD_State <=  2;
        end if;
      elsif (PU_State < 0) and (PD_State > 0) then              -- curent state is low
        if (Vin > Vth_R) then                                   -- go to high state
          PU_State <=  2;
          PD_State <= -2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PU_State;
  break on PD_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PU_State use
    when -2 =>                                      -- PU is turning off
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kf2), "HE");
    when -1 =>                                      -- PU is off
      kpu == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PU_on table
    when 1 =>                                       -- PU is on
      kpu == Ktables(Kf2)(Ktables(Kf2)'left);       -- 1st point in PU_off table
    when 2 =>                                       -- PU is tuning on
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when others =>              -- before any process assignments are available
      kpu == 0.0;
  end case;

  case PD_State use
    when -2 =>                                      -- PD is turning off
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kr2), "HE");
    when -1 =>                                      -- PD is off
      kpd == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PD_on table
    when 1 =>                                       -- PD is on
      kpd == Ktables(Kr2)(Ktables(Kr2)'left);       -- 1st point in PD_off table
    when 2 =>                                       -- PD is tuning on
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when others =>              -- before any process assignments are available
      kpd == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_2EQ2UK;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_IO (PU_ref, PD_ref, Pad, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS I/O buffer. The effects modeled are:
--              - Pullup, pulldown, and clamp I/V curves with IBIS 2.1
--                waveform-based pullup/pulldown turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--              - Threshold based received logic detection.
--
--              The Vth_R and Vth_F parameters control how the In_D and
--              En_D terminals behave. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D and En_D terminals are
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--              For DC analyses the DC_threshold parameter is used, since
--              the hysteresis implemented using Vth_R and Vth_F is not useful
--              in DC analysis. This defaults to the mid-point between
--              Vth_R and Vth_F.
--
--              When En_D is low the Pad terminal goes to high impedance.
--              The voltage at the Pad terminal is monitored and
--              a corresponding output voltage is produced at the Rcv_D
--              terminal. The Vinl and Vinh parameters are used as voltage
--              thresholds for controlling the core-level output data
--              at the Rcv_D terminal. When the Pad voltage is above
--              Vinh the Rcv_D voltage is 1.0. When the Pad voltage is
--              below Vinl the Rcv_D voltage is 0.0. When the Pad voltage
--              is in the middle region the Rcv_D voltage is 0.5.
--
--              When En_D is high the Pad terminal is driven by either the
--              Pullup or Pulldown I/V table, depending on the low/high
--              state of the In_D terminal. State changes at the In_D
--              terminal trigger controlled transitions to the other
--              I/V table, using the waveform and test fixture parameters
--              to guide the transition.
--
--              Clamp, Pullup, and Pulldown currents can be experimentally
--              scaled using the kI_pc, kI_gc, kI_pu, and kI_pd parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  Rcv_D       Received data from buffer to core logic
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpu_length Number of elements in the Ipu_data and Vpu_data arrays
--              Count the number of IBIS [Pullup] entries to get
--              this value.
--              DEFAULT IVpu_length = '4'
--
--  IVpd_length Number of elements in the Ipd_data and Vpd_data arrays
--              Count the number of IBIS [Pulldown] entries to get
--              this value.
--              DEFAULT IVpd_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipu_data[1:IVpu_length]  Array of current values for pullup I/V
--              Extract the IBIS [Pullup] current column to get this.
--              DEFAULT Ipu_data[1:IVpu_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vpu_data[1:IVpu_length]  Array of voltage values for pullup I/V
--              Extract the IBIS [Pullup] voltage column to get this.
--              DEFAULT Vpu_data[1:IVpu_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipd_data[1:IVpd_length]  Array of current values for pulldown I/V
--              Extract the IBIS [Pulldown] current column to get this.
--              DEFAULT Ipd_data[1:IVpd_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpd_data[1:IVpd_length]  Array of voltage values for pulldown I/V
--              Extract the IBIS [Pulldown] voltage column to get this.
--              DEFAULT Vpd_data[1:IVpd_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with V_fixture = 0 to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTr2_length  Number of elements in the Tr2 and Vr2 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTr2_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  VTf2_length  Number of elements in the Tf2 and Vf2 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT VTf2_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vr2_data[1:VTr2_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr2_data[1:VTr2_length] = '{2.50 2.50 5.00 5.00}'
--
--  Tr2_data[1:VTr2_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Tr2_data[1:VTr2_length] = '{0.00 0.50e-9 0.80e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf2_data[1:VTf2_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the higher the higher V_fixture to get this value.
--              DEFAULT Vf2_data[1:VTf2_length] = '{2.50 2.50 0.00 0.00}'
--
--  Tf2_data[1:VTf2_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the higher the higher V_fixture to get this value.
--              DEFAULT Tf2_data[1:VTf2_length] = '{0.00 0.50e-9 0.80e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_r2      V_fixture value for Rising Waveform 2
--              Extract the higher V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r2 = '5.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Vfx_f2      V_fixture value for Falling Waveform 2
--              Extract the lower V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f2 = '0.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with lower V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_r2      R_fixture value for Rising Waveform 2
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with higher V_fixture.
--              DEFAULT Rfx_r2 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with higher V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  Rfx_f2      R_fixture value for Falling Waveform2
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with lower V_fixture.
--              DEFAULT Rfx_f2 = '50.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum delta time between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--  kI_pc       Scaling factor for power clamp current values.
--              DEFAULT kI_pc = '1.0'
--
--  kI_pu       Scaling factor for pullup clamp current values.
--              DEFAULT kI_pu = '1.0'
--
--  kI_pd       Scaling factor for pulldown current values.
--              DEFAULT kI_pd = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              DEFAULT kI_gc = '1.0'
--
--  kt_rise     Scaling factor for [Rising waveform] time values.
--              DEFAULT kt_rise = '1.0 from (0:inf)'
--
--  kt_fall     Scaling factor for [Falling waveform] time values.
--              DEFAULT kt_fall = '1.0 from (0:inf)'
--
--  DC_threshold Input threshold during DC analysis
--              DEFAULT DC_threshold = '(Vth_R+Vth_F)/2.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_IO  #(
--   .C_comp(1.2e-12), .Vinh(2.0), .Vinl(0.8), 
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpu_length(4), .I_pu({ 0.10,  0.00, -0.10, -0.20}),
--                    .V_pu({-5.00,  0.00,  5.00, 10.00}),
--   .IVpd_length(4), .I_pd({-0.10,  0.00,  0.10,  0.20}),
--                    .V_pd({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTr2_length(6), .Vr2({1.66, 1.66,    2.50,    2.50,    3.33, 3.33}),
--                    .Tr2({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.0e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf2_length(4), .Vf2({1.66, 1.66,    0.00,    0.00}),
--                    .Tf2({0.00, 0.50e-9, 0.80e-9, 3.00e-9}) )
--  B1  (B1_pu, B1_pd, B1_pad, B1_in, B1_en, B1_a, B1_pc, B1_gc);
--
--  VHDL:
--        entity IBIS_IO is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_en, B1_a, B1_pc, B1_gc : electrical);
--        end entity IBIS_IO;
--        architecture IBIS_SIM of IBIS_IO is
--        begin
--          generic map ( C_comp => 1.2e-12,
--          Vinh => 2.0, Vinl => 0.8, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_IO is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vr2_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf2_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_r2 : real := 5.0;
           Vfx_f1 : real := 5.0;
           Vfx_f2 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_r2 : real := 50.0;
           Rfx_f1 : real := 50.0;
           Rfx_f2 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;
--           terminal Rcv_D  : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_IO;
--=============================================================================
architecture IBIS_2EQ2UK of IBIS_IO is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;
--  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  signal    PU_State      : integer := 0;
  signal    PD_State      : integer := 0;
  quantity  kpu           : current := 0.0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vr2_vec     : real_vector := FileRead(DataFile, "Vr2_data", Vr2_data, MaxVt_length);
  constant  Tr2_vec     : real_vector := FileRead(DataFile, "Tr2_data", Tr2_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);
  constant  Vf2_vec     : real_vector := FileRead(DataFile, "Vf2_data", Vf2_data, MaxVt_length);
  constant  Tf2_vec     : real_vector := FileRead(DataFile, "Tf2_data", Tf2_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_r2_val  : real := FileRead(DataFile, "Vfx_r2", Vfx_r2);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);
  constant  Vfx_f2_val  : real := FileRead(DataFile, "Vfx_f2", Vfx_f2);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_r2_val  : real := FileRead(DataFile, "Rfx_r2", Rfx_r2);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
  constant  Rfx_f2_val  : real := FileRead(DataFile, "Rfx_f2", Rfx_f2);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tr2_vec, Tf1_vec, Tf2_vec);

  type k_elements    is (Tcm, Kr1, Kr2, Kf1, Kf2);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;
  
                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_r2   : real_vector;
                  T_r2   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;
                  V_f2   : real_vector;
                  T_f2   : real_vector;
  
                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;
  
                  V_fx_r1  : real;
                  V_fx_r2  : real;
                  V_fx_f1  : real;
                  V_fx_f2  : real;
  
                  R_fx_r1  : real;
                  R_fx_r2  : real;
                  R_fx_f1  : real;
                  R_fx_f2  : real;
  
                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;
  
    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vr2cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf2cm : real_vector(1 to 3) := (others => 0.0);
  
    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_r2 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;
    variable dVwfm_f2 : real := 0.0;
  
    variable Iout1 : real := 0.0;
    variable Iout2 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipc2  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipu2  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Ipd2  : real := 0.0;
    variable Igc1  : real := 0.0;
    variable Igc2  : real := 0.0;
  
    variable Term1 : real := 0.0;
    variable Term2 : real := 0.0;
    variable Den   : real := 1.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_r2, T_f1, T_f2);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vr2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r2, V_r2, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");
    Vf2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f2, V_f2, "HE");
  
    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vr2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r2, V_r2, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
        Vf2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f2, V_f2, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " R2: " & real'image(Vr2cm(2)) &
      --       " F1: " & real'image(Vf1cm(2)) &
      --       " F2: " & real'image(Vf2cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_r2 := 0.0;
        dVwfm_f1 := 0.0;
        dVwfm_f2 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_r2 := ((Vr2cm(2) - Vr2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr2cm(3) - Vr2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f2 := ((Vf2cm(2) - Vf2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf2cm(3) - Vf2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dR2: " & real'image(dVwfm_r2) &
      --       " dF1: " & real'image(dVwfm_f1) &
      --       " dF2: " & real'image(dVwfm_f2);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 Kr2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;
      Iout2 := ((Vr2cm(2) - V_fx_r2) / R_fx_r2) + C_comp_total * dVwfm_r2;
  
      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vr2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vr2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vr2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vr2cm(2) - V_gc_ref, V_gc, I_gc, "HE");
  
      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);
  
      -- Since these are rising waveforms
      -- Kr1 == Kpu_on and Kr2 == Kpd_off
      K_tables(Kr1)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      K_tables(Kr2)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 Kf2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;
      Iout2 := ((Vf2cm(2) - V_fx_f2) / R_fx_f2) + C_comp_total * dVwfm_f2;
  
      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vf2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vf2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vf2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vf2cm(2) - V_gc_ref, V_gc, I_gc, "HE");
  
      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);
  
      -- Since these are falling waveforms
      -- Kf1 == Kpd_on and Kf2 == Kpu_off
      K_tables(Kf1)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      K_tables(Kf2)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kr2: " & real'image(K_tables(Kr2)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i)) &
      --       " Kf2: " & real'image(K_tables(Kf2)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vr2cm(1) := Vr2cm(2);
      Vf1cm(1) := Vf1cm(2);
      Vf2cm(1) := Vf2cm(2);
  
      Vr1cm(2) := Vr1cm(3);
      Vr2cm(2) := Vr2cm(3);
      Vf1cm(2) := Vf1cm(3);
      Vf2cm(2) := Vf2cm(3);
  
    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vr2_vec, Tr2_vec,
                                            Vf1_vec, Tf1_vec,
                                            Vf2_vec, Tf2_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_r2_val,
                                            Vfx_f1_val,
                                            Vfx_f2_val,

                                            Rfx_r1_val,
                                            Rfx_r2_val,
                                            Rfx_f1_val,
                                            Rfx_f2_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F),
                                Ven'above(Vth_R), Ven'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Ven > DC_threshold) and (Vin > DC_threshold) then     -- high state
        PU_State <=  1;
        PD_State <= -1;
      elsif (Ven > DC_threshold) and (Vin <= DC_threshold) then -- low state
        PU_State <= -1;
        PD_State <=  1;
      else                                                      -- 3-state
        PU_State <= -1;
        PD_State <= -1;
      end if;
    else
      if (PU_State < 0) and (PD_State < 0) then                 -- curent state is 3-state
        if (Ven > Vth_R) and (Vin > DC_threshold) then          -- go to high state
          PU_State <=  2;
          PD_State <= -1;
        elsif (Ven > Vth_R) and (Vin <= DC_threshold) then      -- go to low state
          PU_State <= -1;
          PD_State <=  2;
        end if;
      elsif (PU_State > 0) and (PD_State < 0) then              -- curent state is high
        if (Ven > Vth_R) and (Vin < Vth_F) then                 -- go to low state
          PU_State <= -2;
          PD_State <=  2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -2;
          PD_State <= -1;
        end if;
      elsif (PU_State < 0) and (PD_State > 0) then              -- curent state is low
        if (Ven > Vth_R) and (Vin > Vth_R) then                 -- go to high state
          PU_State <=  2;
          PD_State <= -2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -1;
          PD_State <= -2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PU_State;
  break on PD_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PU_State use
    when -2 =>                                      -- PU is turning off
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kf2), "HE");
    when -1 =>                                      -- PU is off
      kpu == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PU_on table
    when 1 =>                                       -- PU is on
      kpu == Ktables(Kf2)(Ktables(Kf2)'left);       -- 1st point in PU_off table
    when 2 =>                                       -- PU is tuning on
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when others =>              -- before any process assignments are available
      kpu == 0.0;
  end case;

  case PD_State use
    when -2 =>                                      -- PD is turning off
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kr2), "HE");
    when -1 =>                                      -- PD is off
      kpd == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PD_on table
    when 1 =>                                       -- PD is on
      kpd == Ktables(Kr2)(Ktables(Kr2)'left);       -- 1st point in PD_off table
    when 2 =>                                       -- PD is tuning on
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when others =>              -- before any process assignments are available
      kpd == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc   == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu   == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd   ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc   ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
--  if        Vgc'above(Vinh_val)  use Vrcv == 1.0;
--  elsif not(Vgc'above(Vinl_val)) use Vrcv == 0.0;
--  else                               Vrcv == 0.5;
--  end use;
--  break on Vgc'above(Vinh_val), Vgc'above(Vinl_val);
--=============================================================================
end architecture IBIS_2EQ2UK;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_3STATE (PU_ref, PD_ref, Pad, In_D, En_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS tri-stateable Output buffer. The effects modeled are:
--              - Pullup, pulldown, and clamp I/V curves with IBIS 2.1
--                waveform-based pullup/pulldown turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--
--              The Vth_R and Vth_F parameters control how the In_D and
--              En_D terminals behave. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D and En_D terminals are
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--              For DC analyses the DC_threshold parameter is used, since
--              the hysteresis implemented using Vth_R and Vth_F is not useful
--              in DC analysis. This defaults to the mid-point between
--              Vth_R and Vth_F.
--
--              When En_D is low the Pad terminal goes to high impedance.
--
--              Clamp, Pullup, and Pulldown currents can be experimentally
--              scaled using the kI_pc, kI_gc, kI_pu, and kI_pd parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpu_length Number of elements in the Ipu_data and Vpu_data arrays
--              Count the number of IBIS [Pullup] entries to get
--              this value.
--              DEFAULT IVpu_length = '4'
--
--  IVpd_length Number of elements in the Ipd_data and Vpd_data arrays
--              Count the number of IBIS [Pulldown] entries to get
--              this value.
--              DEFAULT IVpd_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipu_data[1:IVpu_length]  Array of current values for pullup I/V
--              Extract the IBIS [Pullup] current column to get this.
--              DEFAULT Ipu_data[1:IVpu_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vpu_data[1:IVpu_length]  Array of voltage values for pullup I/V
--              Extract the IBIS [Pullup] voltage column to get this.
--              DEFAULT Vpu_data[1:IVpu_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipd_data[1:IVpd_length]  Array of current values for pulldown I/V
--              Extract the IBIS [Pulldown] current column to get this.
--              DEFAULT Ipd_data[1:IVpd_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpd_data[1:IVpd_length]  Array of voltage values for pulldown I/V
--              Extract the IBIS [Pulldown] voltage column to get this.
--              DEFAULT Vpd_data[1:IVpd_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with V_fixture = 0 to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTr2_length  Number of elements in the Tr2 and Vr2 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTr2_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  VTf2_length  Number of elements in the Tf2 and Vf2 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT VTf2_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vr2_data[1:VTr2_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr2_data[1:VTr2_length] = '{2.50 2.50 5.00 5.00}'
--
--  Tr2_data[1:VTr2_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Tr2_data[1:VTr2_length] = '{0.00 0.50e-9 0.80e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf2_data[1:VTf2_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the higher the higher V_fixture to get this value.
--              DEFAULT Vf2_data[1:VTf2_length] = '{2.50 2.50 0.00 0.00}'
--
--  Tf2_data[1:VTf2_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the higher the higher V_fixture to get this value.
--              DEFAULT Tf2_data[1:VTf2_length] = '{0.00 0.50e-9 0.80e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_r2      V_fixture value for Rising Waveform 2
--              Extract the higher V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r2 = '5.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Vfx_f2      V_fixture value for Falling Waveform 2
--              Extract the lower V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f2 = '0.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with lower V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_r2      R_fixture value for Rising Waveform 2
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with higher V_fixture.
--              DEFAULT Rfx_r2 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with higher V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  Rfx_f2      R_fixture value for Falling Waveform2
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with lower V_fixture.
--              DEFAULT Rfx_f2 = '50.0 from (0:inf)'
--
--  kI_pc       Scaling factor for power clamp current values.
--              DEFAULT kI_pc = '1.0'
--
--  kI_pu       Scaling factor for pullup clamp current values.
--              DEFAULT kI_pu = '1.0'
--
--  kI_pd       Scaling factor for pulldown current values.
--              DEFAULT kI_pd = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              DEFAULT kI_gc = '1.0'
--
--  kt_rise     Scaling factor for [Rising waveform] time values.
--              DEFAULT kt_rise = '1.0 from (0:inf)'
--
--  kt_fall     Scaling factor for [Falling waveform] time values.
--              DEFAULT kt_fall = '1.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum dt between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_3STATE  #(
--   .C_comp(1.2e-12),
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpu_length(4), .I_pu({ 0.10,  0.00, -0.10, -0.20}),
--                    .V_pu({-5.00,  0.00,  5.00, 10.00}),
--   .IVpd_length(4), .I_pd({-0.10,  0.00,  0.10,  0.20}),
--                    .V_pd({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTr2_length(6), .Vr2({1.66, 1.66,    2.50,    2.50,    3.33, 3.33}),
--                    .Tr2({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.0e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf2_length(4), .Vf2({1.66, 1.66,    0.00,    0.00}),
--                    .Tf2({0.00, 0.50e-9, 0.80e-9, 3.00e-9}) )
--  B1  (B1_pu,  B1_pd,  B1_pad, B1_in,  B1_pc,    B1_gc);
--
--  VHDL:
--        entity IBIS_3STATE is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_pc, B1_gc : electrical);
--        end entity IBIS_3STATE;
--        architecture IBIS_SIM of IBIS_3STATE is
--        begin
--          generic map ( C_comp => 1.2e-12, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_3STATE is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vr2_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf2_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf2_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_r2 : real := 5.0;
           Vfx_f1 : real := 5.0;
           Vfx_f2 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_r2 : real := 50.0;
           Rfx_f1 : real := 50.0;
           Rfx_f2 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_3STATE;
--=============================================================================
architecture IBIS_2EQ2UK of IBIS_3STATE is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;

  signal    PU_State      : integer := 0;
  signal    PD_State      : integer := 0;
  quantity  kpu           : current := 0.0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vr2_vec     : real_vector := FileRead(DataFile, "Vr2_data", Vr2_data, MaxVt_length);
  constant  Tr2_vec     : real_vector := FileRead(DataFile, "Tr2_data", Tr2_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);
  constant  Vf2_vec     : real_vector := FileRead(DataFile, "Vf2_data", Vf2_data, MaxVt_length);
  constant  Tf2_vec     : real_vector := FileRead(DataFile, "Tf2_data", Tf2_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_r2_val  : real := FileRead(DataFile, "Vfx_r2", Vfx_r2);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);
  constant  Vfx_f2_val  : real := FileRead(DataFile, "Vfx_f2", Vfx_f2);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_r2_val  : real := FileRead(DataFile, "Rfx_r2", Rfx_r2);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
  constant  Rfx_f2_val  : real := FileRead(DataFile, "Rfx_f2", Rfx_f2);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tr2_vec, Tf1_vec, Tf2_vec);

  type k_elements    is (Tcm, Kr1, Kr2, Kf1, Kf2);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_r2   : real_vector;
                  T_r2   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;
                  V_f2   : real_vector;
                  T_f2   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_r2  : real;
                  V_fx_f1  : real;
                  V_fx_f2  : real;

                  R_fx_r1  : real;
                  R_fx_r2  : real;
                  R_fx_f1  : real;
                  R_fx_f2  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vr2cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf2cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_r2 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;
    variable dVwfm_f2 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Iout2 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipc2  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipu2  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Ipd2  : real := 0.0;
    variable Igc1  : real := 0.0;
    variable Igc2  : real := 0.0;

    variable Term1 : real := 0.0;
    variable Term2 : real := 0.0;
    variable Den   : real := 1.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_r2, T_f1, T_f2);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vr2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r2, V_r2, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");
    Vf2cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f2, V_f2, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vr2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r2, V_r2, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
        Vf2cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f2, V_f2, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " R2: " & real'image(Vr2cm(2)) &
      --       " F1: " & real'image(Vf1cm(2)) &
      --       " F2: " & real'image(Vf2cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_r2 := 0.0;
        dVwfm_f1 := 0.0;
        dVwfm_f2 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_r2 := ((Vr2cm(2) - Vr2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr2cm(3) - Vr2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f2 := ((Vf2cm(2) - Vf2cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf2cm(3) - Vf2cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dR2: " & real'image(dVwfm_r2) &
      --       " dF1: " & real'image(dVwfm_f1) &
      --       " dF2: " & real'image(dVwfm_f2);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 Kr2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;
      Iout2 := ((Vr2cm(2) - V_fx_r2) / R_fx_r2) + C_comp_total * dVwfm_r2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vr2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vr2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vr2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vr2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on and Kr2 == Kpd_off
      K_tables(Kr1)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      K_tables(Kr2)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 Kf2 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;
      Iout2 := ((Vf2cm(2) - V_fx_f2) / R_fx_f2) + C_comp_total * dVwfm_f2;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipc2  := -1.0 * Lookup(V_pc_ref - Vf2cm(2), V_pc, I_pc, "HE");
      Ipu1  := -1.0 * Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Ipu2  := -1.0 * Lookup(V_pu_ref - Vf2cm(2), V_pu, I_pu, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Ipd2  :=        Lookup(Vf2cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");
      Igc2  :=        Lookup(Vf2cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      Term1 := Iout1 + Igc1 - Ipc1;
      Term2 := Ipc2  - Igc2 - Iout2;
      Den   := (Ipu1 * Ipd2) - (Ipd1 * Ipu2);

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on and Kf2 == Kpu_off
      K_tables(Kf1)(i) := (Ipu2*Term1 + Ipu1*Term2) / Den;
      K_tables(Kf2)(i) := (Ipd2*Term1 + Ipd1*Term2) / Den;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kr2: " & real'image(K_tables(Kr2)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i)) &
      --       " Kf2: " & real'image(K_tables(Kf2)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vr2cm(1) := Vr2cm(2);
      Vf1cm(1) := Vf1cm(2);
      Vf2cm(1) := Vf2cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vr2cm(2) := Vr2cm(3);
      Vf1cm(2) := Vf1cm(3);
      Vf2cm(2) := Vf2cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vr2_vec, Tr2_vec,
                                            Vf1_vec, Tf1_vec,
                                            Vf2_vec, Tf2_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_r2_val,
                                            Vfx_f1_val,
                                            Vfx_f2_val,

                                            Rfx_r1_val,
                                            Rfx_r2_val,
                                            Rfx_f1_val,
                                            Rfx_f2_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F),
                                Ven'above(Vth_R), Ven'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Ven > DC_threshold) and (Vin > DC_threshold) then     -- high state
        PU_State <=  1;
        PD_State <= -1;
      elsif (Ven > DC_threshold) and (Vin <= DC_threshold) then -- low state
        PU_State <= -1;
        PD_State <=  1;
      else                                                      -- 3-state
        PU_State <= -1;
        PD_State <= -1;
      end if;
    else
      if (PU_State < 0) and (PD_State < 0) then                 -- curent state is 3-state
        if (Ven > Vth_R) and (Vin > DC_threshold) then          -- go to high state
          PU_State <=  2;
          PD_State <= -1;
        elsif (Ven > Vth_R) and (Vin <= DC_threshold) then      -- go to low state
          PU_State <= -1;
          PD_State <=  2;
        end if;
      elsif (PU_State > 0) and (PD_State < 0) then              -- curent state is high
        if (Ven > Vth_R) and (Vin < Vth_F) then                 -- go to low state
          PU_State <= -2;
          PD_State <=  2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -2;
          PD_State <= -1;
        end if;
      elsif (PU_State < 0) and (PD_State > 0) then              -- curent state is low
        if (Ven > Vth_R) and (Vin > Vth_R) then                 -- go to high state
          PU_State <=  2;
          PD_State <= -2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -1;
          PD_State <= -2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PU_State;
  break on PD_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PU_State use
    when -2 =>                                      -- PU is turning off
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kf2), "HE");
    when -1 =>                                      -- PU is off
      kpu == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PU_on table
    when 1 =>                                       -- PU is on
      kpu == Ktables(Kf2)(Ktables(Kf2)'left);       -- 1st point in PU_off table
    when 2 =>                                       -- PU is tuning on
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when others =>              -- before any process assignments are available
      kpu == 0.0;
  end case;

  case PD_State use
    when -2 =>                                      -- PD is turning off
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kr2), "HE");
    when -1 =>                                      -- PD is off
      kpd == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PD_on table
    when 1 =>                                       -- PD is on
      kpd == Ktables(Kr2)(Ktables(Kr2)'left);       -- 1st point in PD_off table
    when 2 =>                                       -- PD is tuning on
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when others =>              -- before any process assignments are available
      kpd == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_2EQ2UK;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_OPENSINK (PU_ref, PD_ref, Pad, In_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS open-sink output buffer, with no pullup I/V.
--              The effects modeled are:
--              - Pulldown and clamp I/V curves with IBIS 2.1
--                waveform-based pulldown turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--                This can distribute current to the PU_ref terminal
--                even though there is no pullup I/V.
--
--              The Vth_R and Vth_F parameters control how the In_D 
--              terminal behaves. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D terminal is
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--
--              The Pad terminal is driven by the
--              Pulldown I/V table, depending on the low/high
--              state of the In_D terminal. State changes at the In_D
--              terminal trigger controlled on/off transitions for the
--              I/V table, using the waveform and test fixture parameters
--              to guide the transition.
--
--              Only one waveform pair is supported by this model instead
--              of the usual two pairs for push-pull buffers. The V_fixture
--              values should be at the power voltage.
--
--              Clamp and Pulldown currents can be experimentally
--              scaled using the kI_pc, kI_gc, and kI_pd parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpd_length Number of elements in the Ipd_data and Vpd_data arrays
--              Count the number of IBIS [Pulldown] entries to get
--              this value.
--              DEFAULT IVpd_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipd_data[1:IVpd_length]  Array of current values for pulldown I/V
--              Extract the IBIS [Pulldown] current column to get this.
--              DEFAULT Ipd_data[1:IVpd_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpd_data[1:IVpd_length]  Array of voltage values for pulldown I/V
--              Extract the IBIS [Pulldown] voltage column to get this.
--              DEFAULT Vpd_data[1:IVpd_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with V_fixture = 0 to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with higher V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with higher V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  kI_pc       Scaling factor for power clamp current values.
--              DEFAULT kI_pc = '1.0'
--
--  kI_pd       Scaling factor for pulldown current values.
--              DEFAULT kI_pd = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              DEFAULT kI_gc = '1.0'
--
--  kt_rise     Scaling factor for [Rising waveform] time values.
--              DEFAULT kt_rise = '1.0 from (0:inf)'
--
--  kt_fall     Scaling factor for [Falling waveform] time values.
--              DEFAULT kt_fall = '1.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum dt between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_OPENSINK  #(
--   .C_comp(1.2e-12),
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpd_length(4), .I_pd({-0.10,  0.00,  0.10,  0.20}),
--                    .V_pd({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}))
--  B1  (B1_pu,  B1_pd,  B1_pad, B1_in,  B1_pc,    B1_gc);
--
--  VHDL:
--        entity IBIS_OPENSINK is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_pc, B1_gc : electrical);
--        end entity IBIS_OPENSINK;
--        architecture IBIS_SIM of IBIS_OPENSINK is
--        begin
--          generic map ( C_comp => 1.2e-12, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_OPENSINK is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 5.0;
           Vfx_f1 : real := 5.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_OPENSINK;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_OPENSINK is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;

  signal    PD_State      : integer := 0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpd_off
      if (Ipd1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on
      if (Ipd1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Vin > DC_threshold) then                              -- high state
        PD_State <= -1;
      else                                                      -- low state
        PD_State <=  1;
      end if;
    else
      if (PD_State < 0) then                                    -- curent state is high
        if (Vin < Vth_F) then                                   -- go to low state
          PD_State <=  2;
        end if;
      elsif (PD_State > 0) then                                 -- curent state is low
        if (Vin > Vth_R) then                                   -- go to high state
          PD_State <= -2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PD_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PD_State use
    when -2 =>                                      -- PD is turning off
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when -1 =>                                      -- PD is off
      kpd == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PD_on table
    when 1 =>                                       -- PD is on
      kpd == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PD_off table
    when 2 =>                                       -- PD is tuning on
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when others =>              -- before any process assignments are available
      kpd == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu ==  0.0                                              ;-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_IO_OPENSINK (PU_ref, PD_ref, Pad, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS open-sink I/O buffer, with no pullup I/V.
--              The effects modeled are:
--              - Pulldown and clamp I/V curves with IBIS 2.1
--                waveform-based pullup/pulldown turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--                This can distribute current to the PU_ref terminal
--                even though there is no pullup I/V.
--              - Threshold based received logic detection.
--
--              The Vth_R and Vth_F parameters control how the In_D and
--              En_D terminals behave. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D and En_D terminals are
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--              For DC analyses the DC_threshold parameter is used, since
--              the hysteresis implemented using Vth_R and Vth_F is not useful
--              in DC analysis. This defaults to the mid-point between
--              Vth_R and Vth_F.
--
--              The Pad terminal is driven by the
--              Pulldown I/V table, depending on the low/high
--              state of the In_D terminal. State changes at the In_D
--              terminal trigger controlled on/off transitions of the
--              I/V table, using the waveform and test fixture parameters
--              to guide the transition.
--
--              Only one waveform pair is supported by this model instead
--              of the usual two pairs for push-pull buffers. The V_fixture
--              values should be at the power voltage.
--
--              Clamp and Pulldown currents can be experimentally
--              scaled using the kI_pc, kI_gc, and kI_pd parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  Rcv_D       Received data from buffer to core logic
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpd_length Number of elements in the Ipd_data and Vpd_data arrays
--              Count the number of IBIS [Pulldown] entries to get
--              this value.
--              DEFAULT IVpd_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipd_data[1:IVpd_length]  Array of current values for pulldown I/V
--              Extract the IBIS [Pulldown] current column to get this.
--              DEFAULT Ipd_data[1:IVpd_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpd_data[1:IVpd_length]  Array of voltage values for pulldown I/V
--              Extract the IBIS [Pulldown] voltage column to get this.
--              DEFAULT Vpd_data[1:IVpd_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the higher the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the higher V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the higher V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with higher V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with higher V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum delta time between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--  kI_pc       Scaling factor for power clamp current values.
--              DEFAULT kI_pc = '1.0'
--
--  kI_pd       Scaling factor for pulldown current values.
--              DEFAULT kI_pd = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              DEFAULT kI_gc = '1.0'
--
--  kt_rise     Scaling factor for [Rising waveform] time values.
--              DEFAULT kt_rise = '1.0 from (0:inf)'
--
--  kt_fall     Scaling factor for [Falling waveform] time values.
--              DEFAULT kt_fall = '1.0 from (0:inf)'
--
--  DC_threshold Input threshold during DC analysis
--              DEFAULT DC_threshold = '(Vth_R+Vth_F)/2.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_IO_OPENSINK  #(
--   .C_comp(1.2e-12), .Vinh(2.0), .Vinl(0.8), 
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpd_length(4), .I_pd({-0.10,  0.00,  0.10,  0.20}),
--                    .V_pd({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}))
--  B1  (B1_pu, B1_pd, B1_pad, B1_in, B1_en, B1_a, B1_pc, B1_gc);
--
--  VHDL:
--        entity IBIS_IO_OPENSINK is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_en, B1_a, B1_pc, B1_gc : electrical);
--        end entity IBIS_IO_OPENSINK;
--        architecture IBIS_SIM of IBIS_IO_OPENSINK is
--        begin
--          generic map ( C_comp => 1.2e-12,
--          Vinh => 2.0, Vinl => 0.8, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_IO_OPENSINK is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpd_ref : real := 0.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipd_data : real_vector := (-0.10,  0.00,  0.10,  0.20);
           Vpd_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (2.50,  2.50,     5.00,     5.00);
           Tr1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           Vf1_data : real_vector := (5.00,  5.00,     2.50,     2.50);
           Tf1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 5.0;
           Vfx_f1 : real := 5.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pd        : real    := 1.0;     -- PD current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;
--           terminal Rcv_D  : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_IO_OPENSINK;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_IO_OPENSINK is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;
--  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  signal    PU_State      : integer := 0;
  signal    PD_State      : integer := 0;
  quantity  kpd           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpd_ref_val : real := FileRead(DataFile, "Vpd_ref", Vpd_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipd_vec     : real_vector := FileRead(DataFile, "Ipd_data", Ipd_data, MaxIV_length);
  constant  Vpd_vec     : real_vector := FileRead(DataFile, "Vpd_data", Vpd_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pd   : real_vector;
                  V_pd   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pd_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vr1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpd_off
      if (Ipd1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipd1  :=        Lookup(Vf1cm(2) - V_pd_ref, V_pd, I_pd, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpd_on
      if (Ipd1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipd1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipd_vec, Vpd_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpd_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F),
                                Ven'above(Vth_R), Ven'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Ven > DC_threshold) and (Vin > DC_threshold) then     -- high state
        PU_State <=  1;
        PD_State <= -1;
      elsif (Ven > DC_threshold) and (Vin <= DC_threshold) then -- low state
        PU_State <= -1;
        PD_State <=  1;
      else                                                      -- 3-state
        PU_State <= -1;
        PD_State <= -1;
      end if;
    else
      if (PU_State < 0) and (PD_State < 0) then                 -- curent state is 3-state
        if (Ven > Vth_R) and (Vin > DC_threshold) then          -- go to high state
          PU_State <=  2;
          PD_State <= -1;
        elsif (Ven > Vth_R) and (Vin <= DC_threshold) then      -- go to low state
          PU_State <= -1;
          PD_State <=  2;
        end if;
      elsif (PU_State > 0) and (PD_State < 0) then              -- curent state is high
        if (Ven > Vth_R) and (Vin < Vth_F) then                 -- go to low state
          PU_State <= -2;
          PD_State <=  2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -2;
          PD_State <= -1;
        end if;
      elsif (PU_State < 0) and (PD_State > 0) then              -- curent state is low
        if (Ven > Vth_R) and (Vin > Vth_R) then                 -- go to high state
          PU_State <=  2;
          PD_State <= -2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -1;
          PD_State <= -2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PD_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PD_State use
    when -2 =>                                      -- PD is turning off
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when -1 =>                                      -- PD is off
      kpd == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PD_on table
    when 1 =>                                       -- PD is on
      kpd == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PD_off table
    when 2 =>                                       -- PD is tuning on
      kpd == Lookup(PD_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when others =>              -- before any process assignments are available
      kpd == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu ==  0.0                                              ;-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==      kI_pd*kpd*Lookup(Vpd, Vpd_vec, Ipd_vec, "SE");-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
--  if        Vgc'above(Vinh_val)  use Vrcv == 1.0;
--  elsif not(Vgc'above(Vinl_val)) use Vrcv == 0.0;
--  else                               Vrcv == 0.5;
--  end use;
--  break on Vgc'above(Vinh_val), Vgc'above(Vinl_val);
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_OPENSOURCE (PU_ref, PD_ref, Pad, In_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS open-source buffer, with no pulldown I/V.
--              The effects modeled are:
--              - Pulldown and clamp I/V curves with IBIS 2.1
--                waveform-based pullup turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--
--              The Vth_R and Vth_F parameters control how the In_D 
--              terminal behaves. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D terminal is
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--
--              The Pad terminal is driven by the
--              Pullup I/V table, depending on the low/high
--              state of the In_D terminal. State changes at the In_D
--              terminal trigger controlled on/off transitions of the
--              I/V table, using the waveform and test fixture parameters
--              to guide the transition.
--
--              Clamp and Pullup currents can be experimentally
--              scaled using the kI_pc, kI_gc, and kI_pu parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpu_length Number of elements in the Ipu_data and Vpu_data arrays
--              Count the number of IBIS [Pullup] entries to get
--              this value.
--              DEFAULT IVpu_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipu_data[1:IVpu_length]  Array of current values for pullup I/V
--              Extract the IBIS [Pullup] current column to get this.
--              DEFAULT Ipu_data[1:IVpu_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vpu_data[1:IVpu_length]  Array of voltage values for pullup I/V
--              Extract the IBIS [Pullup] voltage column to get this.
--              DEFAULT Vpu_data[1:IVpu_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with lower V_fixture to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with lower V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the lower the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with lower V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with lower V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum dt between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_OPENSOURCE  #(
--   .C_comp(1.2e-12),
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpu_length(4), .I_pu({ 0.10,  0.00, -0.10, -0.20}),
--                    .V_pu({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}))
--  B1  (B1_pu,  B1_pd,  B1_pad, B1_in,  B1_pc,    B1_gc);
--
--  VHDL:
--        entity IBIS_OPENSOURCE is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_pc, B1_gc : electrical);
--        end entity IBIS_OPENSOURCE;
--        architecture IBIS_SIM of IBIS_OPENSOURCE is
--        begin
--          generic map ( C_comp => 1.2e-12, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_OPENSOURCE is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf1_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_f1 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_OPENSOURCE;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_OPENSOURCE is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;

  signal    PU_State      : integer := 0;
  quantity  kpu           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on
      if (Ipu1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpu_off
      if (Ipu1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Vin > DC_threshold) then                              -- high state
        PU_State <=  1;
      else                                                      -- low state
        PU_State <= -1;
      end if;
    else
      if (PU_State > 0) then                                    -- curent state is high
        if (Vin < Vth_F) then                                   -- go to low state
          PU_State <= -2;
        end if;
      elsif (PU_State < 0) then                                 -- curent state is low
        if (Vin > Vth_R) then                                   -- go to high state
          PU_State <=  2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PU_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PU_State use
    when -2 =>                                      -- PU is turning off
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when -1 =>                                      -- PU is off
      kpu == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PU_on table
    when 1 =>                                       -- PU is on
      kpu == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PU_off table
    when 2 =>                                       -- PU is tuning on
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when others =>              -- before any process assignments are available
      kpu == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==  0.0                                              ;-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
--=============================================================================
--
-- MODULE:      IBIS_IO_OPENSOURCE (PU_ref, PD_ref, Pad, In_D, En_D, Rcv_D, PC_ref, GC_ref)
--
-- DESCRIPTION:
--              An IBIS open-source I/O buffer, with no pulldown I/V.
--              The effects modeled are:
--              - Pullup and clamp I/V curves with IBIS 2.1
--                waveform-based pullup turn-on/off timing.
--              - Die capacitance seen at the Pad terminal, with distribution
--                to separate power/ground terminal per IBIS 4.1.
--              - Threshold based received logic detection.
--
--              The Vth_R and Vth_F parameters control how the In_D and
--              En_D terminals behave. By default these are set to 0.8V
--              and 0.2V, respectively. The In_D and En_D terminals are
--              triggered to the high state when voltage rises above
--              Vth_R and to the low state when voltage falls below Vth_F.
--              For DC analyses the DC_threshold parameter is used, since
--              the hysteresis implemented using Vth_R and Vth_F is not useful
--              in DC analysis. This defaults to the mid-point between
--              Vth_R and Vth_F.
--
--              The Pad terminal is driven by the
--              Pullup I/V table, depending on the low/high
--              state of the In_D terminal. State changes at the In_D
--              terminal trigger controlled on/off transitions of the
--              I/V table, using the waveform and test fixture parameters
--              to guide the transition.
--
--              Clamp and Pullup currents can be experimentally
--              scaled using the kI_pc, kI_gc, and kI_pu parameters.
--              These are not IBIS parameters.
--
--              This IBIS implementation is unable to read IBIS files
--              directly. Therefore it will be necessary to extract
--              appropriate parameters from IBIS models to pass to
--              this model as HDL parameters. IBIS tables are passed
--              as arrays, with the array lengths given as separate
--              parameters. The default parameter values in this model
--              provide a functional example that does not necessarily
--              model any known "real" buffer. Comments below describe
--              where to find values in the IBIS file.
--
--              Note that while IBIS contains typical, minimum, and
--              maximum values for most parameters, this buffer model
--              implementation accepts only a single value for each
--              parameter. It will be necessary to maintain seperate
--              parameter sets for each "corner".
--              
-- TERMINALS:
--
--  PU_ref      Pullup reference voltage supply
--  PD_ref      Pulldown reference voltage supply
--  Pad         Pad terminal providing input to and output from buffer
--  In_D        Core data signal controlling buffer output
--  En_D        Core signal controlling buffer enable state
--  Rcv_D       Received data from buffer to core logic
--  PC_ref      Power clamp reference voltage supply
--  GC_ref      Ground clamp reference voltage supply
--
-- PARAMETERS:
--
--  C_comp      Component die capacitance in farads
--              Extract IBIS C_comp to get this value.
--              DEFAULT C_comp = '5.0p'
--
--  kC_comp_pc  Fraction of C_comp referenced to PC_ref terminal
--              Extract IBIS C_comp_power_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pc = '0.25'
--
--  kC_comp_pu  Fraction of C_comp referenced to PU_ref terminal
--              Extract IBIS C_comp_pullup to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pu = '0.25'
--
--  kC_comp_pd  Fraction of C_comp referenced to PD_ref terminal
--              Extract IBIS C_comp_pulldown to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_pd = '0.25'
--
--  kC_comp_gc  Fraction of C_comp referenced to GC_ref terminal
--              Extract IBIS C_comp_ground_clamp to get this value.
--              The sum of all kC_comp_* values should be 1.0.
--              DEFAULT k_C_comp_gc = '0.25'
--
--  Vinh        Minimum threshold voltage for a logic high to be received
--              Extract IBIS Vinh to get this value.
--              DEFAULT Vinh = '2.0'
--
--  Vinl        Maximum threshold voltage for a logic low to be received
--              Extract IBIS Vinl to get this value.
--              DEFAULT Vinl = '0.8'
--
--  Vpc_ref     Power clamp reference supply voltage
--              Extract IBIS [POWER Clamp Reference] to get this value.
--              Use [Voltage Range] if no [POWER Clamp Reference]
--              DEFAULT Vpc_ref = '5.0'
--
--  Vpu_ref     Pullup reference supply voltage
--              Extract IBIS [Pullup Reference] to get this value.
--              Use [Voltage Range] if no [Pullup Reference]
--              DEFAULT Vpu_ref = '5.0'
--
--  Vpd_ref     Pulldown reference supply voltage
--              Extract IBIS [Pulldown Reference] to get this value.
--              Use 0 if no [Pulldown Reference]
--              DEFAULT Vpd_ref = '0.0'
--
--  Vgc_ref     Ground clamp reference supply voltage
--              Extract IBIS [GND Clamp Reference] to get this value.
--              Use 0 if no [GND Clamp Reference]
--              DEFAULT Vgc_ref = '0.0'
--
--  IVpc_length Number of elements in the Ipc_data and Vpc_data arrays
--              Count the number of IBIS [Power clamp] entries to get
--              this value.
--              DEFAULT IVpc_length = '4'
--
--  IVpu_length Number of elements in the Ipu_data and Vpu_data arrays
--              Count the number of IBIS [Pullup] entries to get
--              this value.
--              DEFAULT IVpu_length = '4'
--
--  IVpd_length Number of elements in the Ipd_data and Vpd_data arrays
--              Count the number of IBIS [Pulldown] entries to get
--              this value.
--              DEFAULT IVpd_length = '4'
--
--  IVgc_length Number of elements in the Igc_data and Vgc_data arrays
--              Count the number of IBIS [GND clamp] entries to get
--              this value.
--              DEFAULT IVgc_length = '4'
--
--  Ipc_data[1:IVpc_length]  Array of current values for power clamp I/V
--              Extract the IBIS [Power clamp] current column to get this.
--              DEFAULT Ipc_data[1:IVpc_length] = '{ 0.08 0.00 0.00 0.00}'
--
--  Vpc_data[1:IVpc_length]  Array of voltage values for power clamp I/V
--              Extract the IBIS [Power clamp] voltage column to get this.
--              DEFAULT Vpc_data[1:IVpc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Ipu_data[1:IVpu_length]  Array of current values for pullup I/V
--              Extract the IBIS [Pullup] current column to get this.
--              DEFAULT Ipu_data[1:IVpu_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vpu_data[1:IVpu_length]  Array of voltage values for pullup I/V
--              Extract the IBIS [Pullup] voltage column to get this.
--              DEFAULT Vpu_data[1:IVpu_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  Igc_data[1:IVgc_length]  Array of current values for ground clamp I/V
--              Extract the IBIS [GND clamp] current column to get this.
--              DEFAULT Igc_data[1:IVgc_length] = '{-0.08 0.00 0.00 0.00}'
--
--  Vgc_data[1:IVgc_length]  Array of voltage values for ground clamp I/V
--              Extract the IBIS [GND clamp] voltage column to get this.
--              DEFAULT Vgc_data[1:IVgc_length] = '{-5.00 -1.00 5.00 10.00}'
--
--  VTr1_length  Number of elements in the Tr1 and Vr1 arrays
--              Count the number of entries in the IBIS [Rising Waveform]
--              with V_fixture = 0 to get this value.
--              DEFAULT VTr1_length = '4'
--
--  VTf1_length  Number of elements in the Tf1 and Vf1 arrays
--              Count the number of entries in the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT VTf1_length = '4'
--
--  Vr1_data[1:VTr1_length]  
--              Extract the voltage column from the IBIS [Rising Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vr1_data[1:VTr1_length] = '{0.00 0.00 2.50 2.50}'
--
--  Tr1_data[1:VTr1_length]  
--              Extract the time column from the IBIS [Rising Waveform]
--              with the lower the lower V_fixture to get this value.
--              DEFAULT Tr1_data[1:VTr1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vf1_data[1:VTf1_length]  
--              Extract the voltage column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Vf1_data[1:VTf1_length] = '{5.00 5.00 2.50 2.50}'
--
--  Tf1_data[1:VTf1_length]  
--              Extract the time column from the IBIS [Falling Waveform]
--              with the lower V_fixture to get this value.
--              DEFAULT Tf1_data[1:VTf1_length] = '{0.00 1.00e-9 2.00e-9 3.00e-9}'
--
--  Vfx_r1      V_fixture value for Rising Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Rising Waveform] set.
--              DEFAULT Vfx_r1 = '0.0'
--
--  Vfx_f1      V_fixture value for Falling Waveform 1
--              Extract the lower V_fixture value from the IBIS
--              [Falling Waveform] set.
--              DEFAULT Vfx_f1 = '5.0'
--
--  Rfx_r1      R_fixture value for Rising Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Rising Waveform] with lower V_fixture.
--              DEFAULT Rfx_r1 = '50.0 from (0:inf)'
--
--  Rfx_f1      R_fixture value for Falling Waveform 1
--              Extract the R_fixture value from the IBIS
--              [Falling Waveform] with lower V_fixture.
--              DEFAULT Rfx_f1 = '50.0 from (0:inf)'
--
--  Vth_R       Input threshold for rising edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the high state
--              when the voltage is above this value.
--              DEFAULT Vth_R = '0.8'
--
--  Vth_F       Input threshold for falling edges
--              Not an IBIS parameter; signals at the In_D and En_D
--              terminals will be considered to be in the low state
--              when the voltage is below this value.
--              DEFAULT Vth_F = '0.2'
--
--  Max_dt      Maximum delta time between simulated time points
--              Not an IBIS parameter; set to desired simulation resolution.
--              DEFAULT Max_dt = '1.0e-12 from (0:inf)'
--
--  kI_pc       Scaling factor for power clamp current values.
--              DEFAULT kI_pc = '1.0'
--
--  kI_pu       Scaling factor for pullup clamp current values.
--              DEFAULT kI_pu = '1.0'
--
--  kI_gc       Scaling factor for ground clamp current values.
--              DEFAULT kI_gc = '1.0'
--
--  kt_rise     Scaling factor for [Rising waveform] time values.
--              DEFAULT kt_rise = '1.0 from (0:inf)'
--
--  kt_fall     Scaling factor for [Falling waveform] time values.
--              DEFAULT kt_fall = '1.0 from (0:inf)'
--
--  DC_threshold Input threshold during DC analysis
--              DEFAULT DC_threshold = '(Vth_R+Vth_F)/2.0'
--
--
-- INSTANCE CALL EXAMPLES:
--
-- VAMS:
--  IBIS_IO_OPENSOURCE  #(
--   .C_comp(1.2e-12), .Vinh(2.0), .Vinl(0.8), 
--   .Vpc_ref(5.0), .Vpu_ref(5.0), .Vpd_ref(0.0), .Vgc_ref(0.0),
--   .IVpc_length(4), .I_pc({ 0.08,  0.00,  0.00,  0.00}),
--                    .V_pc({-5.00, -1.00,  5.00, 10.00}),
--   .IVpu_length(4), .I_pu({ 0.10,  0.00, -0.10, -0.20}),
--                    .V_pu({-5.00,  0.00,  5.00, 10.00}),
--   .IVgc_length(4), .I_gc({-0.10,  0.00,  0.10,  0.20}),
--                    .V_gc({-5.00,  0.00,  5.00, 10.00}),
--   .VTr1_length(4), .Vr1({0.00, 0.00,    1.66,    1.66}),
--                    .Tr1({0.00, 1.00e-9, 2.00e-9, 3.00e-9}),
--   .VTf1_length(6), .Vf1({3.33, 3.33,    2.50,    2.50,    1.66, 1.66}),
--                    .Tf1({0.00, 0.50e-9, 0.80e-9, 1.00e-9, 2.00e-9, 3.00e-9}))
--  B1  (B1_pu, B1_pd, B1_pad, B1_in, B1_en, B1_a, B1_pc, B1_gc);
--
--  VHDL:
--        entity IBIS_IO_OPENSOURCE is
--          port (terminal B1_pu, B1_pd, B1_pad, B1_in,
--          B1_en, B1_a, B1_pc, B1_gc : electrical);
--        end entity IBIS_IO_OPENSOURCE;
--        architecture IBIS_SIM of IBIS_IO_OPENSOURCE is
--        begin
--          generic map ( C_comp => 1.2e-12,
--          Vinh => 2.0, Vinl => 0.8, Vpc_ref => 5.0,
--          Vpu_ref => 5.0, Vpd_ref => 0.0, Vgc_ref => 0.0,
--          DataFile => B1_data.txt )
--        end architecture IBIS_SIM
--=============================================================================
library IEEE;
use IEEE.electrical_systems.all;
use IEEE.math_real.all;

library MacroLib;
use MacroLib.MacroLib_functions.all;
--=============================================================================
entity IBIS_IO_OPENSOURCE is

  generic (C_comp     : real := 5.00e-12;   -- Default C_comp value and
           kC_comp_pc : real := 0.25;       -- splitting coefficients
           kC_comp_pu : real := 0.25;
           kC_comp_pd : real := 0.25;
           kC_comp_gc : real := 0.25;
           --------------------------------------------------------------------
           -- Receiver thresholds
           --------------------------------------------------------------------
           Vinh : real := 2.00;
           Vinl : real := 0.80;
           --------------------------------------------------------------------
           -- [Pullup Reference] and [Pulldown Reference] values
           --------------------------------------------------------------------
           Vpc_ref : real := 5.0;
           Vpu_ref : real := 5.0;
           Vgc_ref : real := 0.0;
           --------------------------------------------------------------------
           -- Vectors of the IV curve tables
           --------------------------------------------------------------------
           Ipc_data : real_vector := ( 0.08,  0.00,  0.00,  0.00);
           Vpc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           Ipu_data : real_vector := ( 0.10,  0.00, -0.10, -0.20);
           Vpu_data : real_vector := (-5.00,  0.00,  5.00, 10.00);
           Igc_data : real_vector := (-0.08,  0.00,  0.00,  0.00);
           Vgc_data : real_vector := (-5.00, -1.00,  5.00, 10.00);
           --------------------------------------------------------------------
           -- Vectors of the Vt curve tables
           --------------------------------------------------------------------
           Vr1_data : real_vector := (0.00,  0.00,     2.50,     2.50);
           Tr1_data : real_vector := (0.00,  1.00e-9,  2.00e-9,  3.00e-9);
           Vf1_data : real_vector := (2.50,  2.50,     0.00,     0.00);
           Tf1_data : real_vector := (0.00,  0.50e-9,  0.80e-9,  3.00e-9);
           --------------------------------------------------------------------
           -- V_fixture and R_fixture values
           --------------------------------------------------------------------
           Vfx_r1 : real := 0.0;
           Vfx_f1 : real := 0.0;

           Rfx_r1 : real := 50.0;
           Rfx_f1 : real := 50.0;
           --------------------------------------------------------------------
           -- Non IBIS parameters
           --------------------------------------------------------------------
           kI_pc        : real    := 1.0;     -- PC current scaling
           kI_pu        : real    := 1.0;     -- PU current scaling
           kI_gc        : real    := 1.0;     -- GC current scaling

           kt_rise      : real    := 1.0;     -- Rising waveform time scaling
           kt_fall      : real    := 1.0;     -- Falling waveform time scaling
           --------------------------------------------------------------------
           DataFile     : string  := "";      -- IBIS parameter file name
           Max_dt       : real    := 1.0e-12; -- Maximum dt between time points
           Vth_R        : real    := 0.8;     -- Input threshold for rising edges
           Vth_F        : real    := 0.2;     -- Input threshold for falling edges
           MaxIV_length : integer := 100;     -- Maximum possible length of IV tables
           MaxVt_length : integer := 1000);   -- Maximum possible length of Vt tables
     --------------------------------------------------------------------------
     port (terminal PU_ref : electrical;
           terminal PD_ref : electrical;

           terminal Pad    : electrical;
           terminal In_D   : electrical;
           terminal En_D   : electrical;
--           terminal Rcv_D  : electrical;

           terminal PC_ref : electrical;
           terminal GC_ref : electrical);

end entity IBIS_IO_OPENSOURCE;
--=============================================================================
architecture IBIS_1EQ1UK of IBIS_IO_OPENSOURCE is

  quantity  Vpc   across  Ipc   through  PC_ref  to  Pad;
  quantity  Vpu   across  Ipu   through  PU_ref  to  Pad;
  quantity  Vpd   across  Ipd   through  Pad     to  PD_ref;
  quantity  Vgc   across  Igc   through  Pad     to  GC_ref;

  quantity  Vin   across                 In_D    to  ELECTRICAL_REF;
  quantity  Ven   across                 En_D    to  ELECTRICAL_REF;
--  quantity  Vrcv  across  Ircv  through  Rcv_D   to  ELECTRICAL_REF;

  signal    PU_State      : integer := 0;
  signal    PD_State      : integer := 0;
  quantity  kpu           : current := 0.0;
  --===========================================================================
  constant  C_comp_val     : real := FileRead(DataFile, "C_comp",     C_comp);
  constant  kC_comp_pc_val : real := FileRead(DataFile, "kC_comp_pc", kC_comp_pc);
  constant  kC_comp_pu_val : real := FileRead(DataFile, "kC_comp_pu", kC_comp_pu);
  constant  kC_comp_pd_val : real := FileRead(DataFile, "kC_comp_pd", kC_comp_pd);
  constant  kC_comp_gc_val : real := FileRead(DataFile, "kC_comp_gc", kC_comp_gc);

  constant  Vinh_val    : real := FileRead(DataFile, "Vinh", Vinh);
  constant  Vinl_val    : real := FileRead(DataFile, "Vinl", Vinl);

  constant  Vpc_ref_val : real := FileRead(DataFile, "Vpc_ref", Vpc_ref);
  constant  Vpu_ref_val : real := FileRead(DataFile, "Vpu_ref", Vpu_ref);
  constant  Vgc_ref_val : real := FileRead(DataFile, "Vgc_ref", Vgc_ref);

  constant  Ipc_vec     : real_vector := FileRead(DataFile, "Ipc_data", Ipc_data, MaxIV_length);
  constant  Vpc_vec     : real_vector := FileRead(DataFile, "Vpc_data", Vpc_data, MaxIV_length);
  constant  Ipu_vec     : real_vector := FileRead(DataFile, "Ipu_data", Ipu_data, MaxIV_length);
  constant  Vpu_vec     : real_vector := FileRead(DataFile, "Vpu_data", Vpu_data, MaxIV_length);
  constant  Igc_vec     : real_vector := FileRead(DataFile, "Igc_data", Igc_data, MaxIV_length);
  constant  Vgc_vec     : real_vector := FileRead(DataFile, "Vgc_data", Vgc_data, MaxIV_length);

  constant  Vr1_vec     : real_vector := FileRead(DataFile, "Vr1_data", Vr1_data, MaxVt_length);
  constant  Tr1_vec     : real_vector := FileRead(DataFile, "Tr1_data", Tr1_data, MaxVt_length, kt_rise);
  constant  Vf1_vec     : real_vector := FileRead(DataFile, "Vf1_data", Vf1_data, MaxVt_length);
  constant  Tf1_vec     : real_vector := FileRead(DataFile, "Tf1_data", Tf1_data, MaxVt_length, kt_fall);

  constant  Vfx_r1_val  : real := FileRead(DataFile, "Vfx_r1", Vfx_r1);
  constant  Vfx_f1_val  : real := FileRead(DataFile, "Vfx_f1", Vfx_f1);

  constant  Rfx_r1_val  : real := FileRead(DataFile, "Rfx_r1", Rfx_r1);
  constant  Rfx_f1_val  : real := FileRead(DataFile, "Rfx_f1", Rfx_f1);
--=============================================================================
  constant CommonLength : integer := FindCommonLength(Max_dt, Tr1_vec, Tf1_vec);

  type k_elements    is (Tcm, Kr1, Kf1);
  type k_tables_type is array(k_elements) of real_vector(1 to CommonLength);
--=============================================================================
  function Coeff (Common_length : integer;
                  Max_dt : real;
                  I_pc   : real_vector;
                  V_pc   : real_vector;
                  I_pu   : real_vector;
                  V_pu   : real_vector;
                  I_gc   : real_vector;
                  V_gc   : real_vector;

                  V_r1   : real_vector;
                  T_r1   : real_vector;
                  V_f1   : real_vector;
                  T_f1   : real_vector;

                  V_pc_ref : real;
                  V_pu_ref : real;
                  V_gc_ref : real;

                  V_fx_r1  : real;
                  V_fx_f1  : real;

                  R_fx_r1  : real;
                  R_fx_f1  : real;

                  C_comp_total : real) return k_tables_type is
  -----------------------------------------------------------------------------
  -- This function converts each Vt table to a corresponding scaling
  -- coefficient table that is used to scale the IV curves with respect
  -- to time.  The C_comp effects in the waveforms are eliminated from the
  -- scaling coefficients, so that the output waveforms will match the Vt
  -- table regardless of the presence of C_comp under identical loading
  -- conditions, i.e. C_comp in the output stage will not derate the
  -- original Vt curves.
  -----------------------------------------------------------------------------
    variable K_tables : k_tables_type;

    variable Vr1cm : real_vector(1 to 3) := (others => 0.0);
    variable Vf1cm : real_vector(1 to 3) := (others => 0.0);

    variable dVwfm_r1 : real := 0.0;
    variable dVwfm_f1 : real := 0.0;

    variable Iout1 : real := 0.0;
    variable Ipc1  : real := 0.0;
    variable Ipu1  : real := 0.0;
    variable Ipd1  : real := 0.0;
    variable Igc1  : real := 0.0;
  -----------------------------------------------------------------------------
  begin
    K_tables(Tcm) := CommonTime(Common_length, Max_dt, T_r1, T_f1);
    --=========================================================================
    -- Calculate the scaling coefficients for each time point along Tcm
    ---------------------------------------------------------------------------
    -- Store the first voltage point in the common waveform variables
    ---------------------------------------------------------------------------
    Vr1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_r1, V_r1, "HE");
    Vf1cm(2) := Lookup(K_tables(Tcm)(K_tables(Tcm)'left), T_f1, V_f1, "HE");

    for i in K_tables(Tcm)'range loop
    --report "Tcm: " & real'image(K_tables(Tcm)(i));  -- For debugging
    ---------------------------------------------------------------------------
    -- Store the next point (for the derivative calculations)
    ---------------------------------------------------------------------------
      if (i < K_tables(Tcm)'right) then
        Vr1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_r1, V_r1, "HE");
        Vf1cm(3) := Lookup(K_tables(Tcm)(i+K_tables(Tcm)'left), T_f1, V_f1, "HE");
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: " & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " R1: " & real'image(Vr1cm(2)) &
      --       " F1: " & real'image(Vf1cm(2));
      -------------------------------------------------------------------------
      -- Calculate the derivative of each waveform for C_comp compensation
      -------------------------------------------------------------------------
      if ((i <= K_tables(Tcm)'left) or (i >= K_tables(Tcm)'right)) then
        dVwfm_r1 := 0.0;          -- Force the end point derivatives to zero
        dVwfm_f1 := 0.0;
      else
        dVwfm_r1 := ((Vr1cm(2) - Vr1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vr1cm(3) - Vr1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
        dVwfm_f1 := ((Vf1cm(2) - Vf1cm(1)) / (K_tables(Tcm)(i)   - K_tables(Tcm)(i-1)) +
                     (Vf1cm(3) - Vf1cm(2)) / (K_tables(Tcm)(i+1) - K_tables(Tcm)(i)))  / 2.0;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " dR1: " & real'image(dVwfm_r1) &
      --       " dF1: " & real'image(dVwfm_f1);
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kr1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vr1cm(2) - V_fx_r1) / R_fx_r1) + C_comp_total * dVwfm_r1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vr1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vr1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vr1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are rising waveforms
      -- Kr1 == Kpu_on
      if (Ipu1 = 0.0) then K_tables(Kr1)(i) := 0.0;
      else                 K_tables(Kr1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      -- Calculate intermediate variables and the Kf1 coefficients
      -------------------------------------------------------------------------
      Iout1 := ((Vf1cm(2) - V_fx_f1) / R_fx_f1) + C_comp_total * dVwfm_f1;

      Ipc1  := -1.0 * Lookup(V_pc_ref - Vf1cm(2), V_pc, I_pc, "HE");
      Ipu1  :=        Lookup(V_pu_ref - Vf1cm(2), V_pu, I_pu, "HE");
      Igc1  :=        Lookup(Vf1cm(2) - V_gc_ref, V_gc, I_gc, "HE");

      -- Since these are falling waveforms
      -- Kf1 == Kpu_off
      if (Ipu1 = 0.0) then K_tables(Kf1)(i) := 0.0;
      else                 K_tables(Kf1)(i) := (Ipc1 - Igc1 - Iout1) / Ipu1;
      end if;
      -------------------------------------------------------------------------
      --report "Tcm: "  & real'image(K_tables(Tcm)(i)) & -- For Debugging
      --       " Kr1: " & real'image(K_tables(Kr1)(i)) &
      --       " Kf1: " & real'image(K_tables(Kf1)(i));
      -------------------------------------------------------------------------
      -- Shift points by one index for next iteration (for C_comp compensation)
      -------------------------------------------------------------------------
      Vr1cm(1) := Vr1cm(2);
      Vf1cm(1) := Vf1cm(2);

      Vr1cm(2) := Vr1cm(3);
      Vf1cm(2) := Vf1cm(3);

    end loop; -- for loop
    ---------------------------------------------------------------------------
    return K_tables;
  end function Coeff;
--=============================================================================
  constant Ktables : k_tables_type := Coeff(CommonLength, Max_dt,

                                            Ipc_vec, Vpc_vec,
                                            Ipu_vec, Vpu_vec,
                                            Igc_vec, Vgc_vec,

                                            Vr1_vec, Tr1_vec,
                                            Vf1_vec, Tf1_vec,

                                            Vpc_ref_val,
                                            Vpu_ref_val,
                                            Vgc_ref_val,

                                            Vfx_r1_val,
                                            Vfx_f1_val,

                                            Rfx_r1_val,
                                            Rfx_f1_val,

                                            C_comp_val);
--=============================================================================
  constant  DC_threshold : real := (Vth_R + Vth_F) / 2.0;
--=============================================================================
begin

  -----------------------------------------------------------------------------
  Threshold_crossing : process (Vin'above(DC_threshold),
                                Vin'above(Vth_R), Vin'above(Vth_F),
                                Ven'above(Vth_R), Ven'above(Vth_F)) is
  -----------------------------------------------------------------------------
  begin
    if domain = quiescent_domain then                           -- during DC analysis
      if (Ven > DC_threshold) and (Vin > DC_threshold) then     -- high state
        PU_State <=  1;
        PD_State <= -1;
      elsif (Ven > DC_threshold) and (Vin <= DC_threshold) then -- low state
        PU_State <= -1;
        PD_State <=  1;
      else                                                      -- 3-state
        PU_State <= -1;
        PD_State <= -1;
      end if;
    else
      if (PU_State < 0) and (PD_State < 0) then                 -- curent state is 3-state
        if (Ven > Vth_R) and (Vin > DC_threshold) then          -- go to high state
          PU_State <=  2;
          PD_State <= -1;
        elsif (Ven > Vth_R) and (Vin <= DC_threshold) then      -- go to low state
          PU_State <= -1;
          PD_State <=  2;
        end if;
      elsif (PU_State > 0) and (PD_State < 0) then              -- curent state is high
        if (Ven > Vth_R) and (Vin < Vth_F) then                 -- go to low state
          PU_State <= -2;
          PD_State <=  2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -2;
          PD_State <= -1;
        end if;
      elsif (PU_State < 0) and (PD_State > 0) then              -- curent state is low
        if (Ven > Vth_R) and (Vin > Vth_R) then                 -- go to high state
          PU_State <=  2;
          PD_State <= -2;
        elsif (Ven < Vth_F) then                                -- go to 3-state
          PU_State <= -1;
          PD_State <= -2;
        end if;
      end if;
    end if;
  end process Threshold_crossing;
  -----------------------------------------------------------------------------
  break on PU_State;
--=============================================================================
  -- This section contains the simultaneous analog equations to find the
  -- appropriate scaling coefficients according to the state the buffer.
  -----------------------------------------------------------------------------

  case PU_State use
    when -2 =>                                      -- PU is turning off
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kf1), "HE");
    when -1 =>                                      -- PU is off
      kpu == Ktables(Kr1)(Ktables(Kr1)'left);       -- 1st point in PU_on table
    when 1 =>                                       -- PU is on
      kpu == Ktables(Kf1)(Ktables(Kf1)'left);       -- 1st point in PU_off table
    when 2 =>                                       -- PU is tuning on
      kpu == Lookup(PU_State'last_event, Ktables(Tcm), Ktables(Kr1), "HE");
    when others =>              -- before any process assignments are available
      kpu == 0.0;
  end case;
--=============================================================================
  -- This section contains the simultaneous analog equations which calculate
  -- the output currents of the IV curves and C_comp capacitors.
  -----------------------------------------------------------------------------
  Ipc == -1.0*kI_pc    *Lookup(Vpc, Vpc_vec, Ipc_vec, "SE");-- + kC_comp_pc_val*C_comp_val*Vpc'dot;
  Ipu == -1.0*kI_pu*kpu*Lookup(Vpu, Vpu_vec, Ipu_vec, "SE");-- + kC_comp_pu_val*C_comp_val*Vpu'dot;
  Ipd ==  0.0                                              ;-- + kC_comp_pd_val*C_comp_val*Vpd'dot;
  Igc ==      kI_gc    *Lookup(Vgc, Vgc_vec, Igc_vec, "SE");-- + kC_comp_gc_val*C_comp_val*Vgc'dot;
  -----------------------------------------------------------------------------
  -- A simple receiver logic
  -----------------------------------------------------------------------------
--  if        Vgc'above(Vinh_val)  use Vrcv == 1.0;
--  elsif not(Vgc'above(Vinl_val)) use Vrcv == 0.0;
--  else                               Vrcv == 0.5;
--  end use;
--  break on Vgc'above(Vinh_val), Vgc'above(Vinl_val);
--=============================================================================
end architecture IBIS_1EQ1UK;
--=============================================================================
